/* 
 * CUAG AutoUmpire System
 * 
 * Copyright (C) 2003/2004 Matthew Johnson, Adam Biltcliffe,
 * Michael Cripps, Martin O'Leary, Edward Allcutt and James Osborn
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * To Contact the authors, please email assassins@srcf.ucam.org
 *
 */

package org.ucam.srcf.assassins;

import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.IOException;
import java.util.Iterator;
import java.util.List;
import java.util.Vector;

import org.ucam.srcf.assassins.exceptions.AssassinSQLException;

/**
 * Class to hold static methods for selecting things from a menu.
 */
public class MenuSelection
{
   static final int rows = 20;
   
   static CanonicalEntity selectEntity(List l) throws AssassinSQLException, IOException
   {
      BufferedReader inp = new BufferedReader(new InputStreamReader(System.in));
      int num = 1;
      while(true)
      {
         for (int i = 0; i < rows && i < l.size(); i++)
         {
            if (Debug.debug) Debug.print(MenuSelection.class, ""+(CanonicalEntity)l.get(num-1));
            System.out.println(num + ": " + ((CanonicalEntity)l.get(num-1)).getPrintable());
            num++;
            if (num >= l.size()+1) { num = 1; }
         }
         System.out.print("Give number (0 to exit) or press enter for more: ");
         String r = inp.readLine();
         try
         {
            int result = Integer.parseInt(r);
            if (result > 0 && result <= l.size())
            {
               return (CanonicalEntity) l.get(result-1);
            }
            if (0 == result) return null;
         }
         catch (NumberFormatException NFe)
         {
         }
      }
   }

   static void askQuestions(Vector v) throws IOException, AssassinSQLException
   {
      Iterator qs = v.iterator();
      BufferedReader in = new BufferedReader(new InputStreamReader(System.in));
      while (qs.hasNext())
      {
         Question q = (Question) qs.next();
         if (q instanceof YesNoQuestion)
         {
            YesNoQuestion ynq = (YesNoQuestion) q;
            System.out.print(q.label);
            System.out.print(ynq.getDefault() ? " (Y/n)" : " (y/N)");
            String s = in.readLine();
            char c = '\n';
            if (null != s && !"".equals(s)) c = s.charAt(0);
            ynq.setAnswer(ynq.getDefault());
            if ('y' == c || 'Y' == c) { ynq.setAnswer(true); }
            if ('n' == c || 'N' == c) { ynq.setAnswer(false); }     
         }
         else if (q instanceof TextQuestion)
         {
            TextQuestion tq = (TextQuestion) q;
            System.out.print(q.label);
            System.out.print(" ("+tq.getDefault()+"): ");
            String r = in.readLine();
            if (null == r || "".equals(r)) { tq.setAnswer(tq.getDefault()); }
            else { tq.setAnswer(r); }
         }
         else if (q instanceof OptionQuestion)
         {	// Blame Mike if this doesn't work
         	OptionQuestion oq = (OptionQuestion) q;
         	System.out.println(oq.label);
         	List l = oq.getOptions();
         	Iterator i = l.iterator();
         	int idx=0;
         	while (i.hasNext()) {
               Object o = i.next();
               String s;
               if (o instanceof CanonicalEntity)
                  s = ((CanonicalEntity) o).getPrintable();
               else if (o instanceof CanonicalViewer)
                  s = ((CanonicalEntity) ((CanonicalViewer) o).getEntity()).getPrintable();
               else
                  s = o.toString();
         		System.out.print(" "+idx+". ");
         		if (idx==oq.getDefault()) {
         			System.out.println("("+s+")");
         		} else {
         			System.out.println(s);
         		}
         		++idx;
         	}
         	String r = in.readLine();
         	if (null == r || "".equals(r)) {
         		oq.setAnswer(oq.getDefault());
         	} else {
         		try {
         			oq.setAnswer(Integer.parseInt(r));
         		} catch (NumberFormatException nfe) {
         			oq.setAnswer(oq.getDefault());
         		}
         	}
         }
      }
   }
}
