package org.ucam.srcf.assassins.strawberry;

import org.ucam.srcf.assassins.AssassinViewer;

import org.ucam.srcf.assassins.exceptions.ConfigFileException;

import java.util.List;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Stack;

/**
 * Class to interpret value of a variable for Scorer system.
 */
class Variable {
	String name=null;

	Variable(String varname) {
		name=new String(varname);
	}

	Variable(char name) {
		this(""+name);
	}
	
	public double evaluate(AssassinViewer killer, AssassinViewer victim, 
			VarTracker vt) 
	{
		switch(name.charAt(0)) {
		/*killer*/
		case 'A':
			if(name.substring(1, 3).equals("is")) {
				throw new RuntimeException("Strawberry: unrecognised variable - \"" + name + "\"");
			} else {
				/*returns 0 if doens't exist*/
				return vt.getIntProperty(killer, name.substring(1));
			}
			
		/*Victim*/
		case 'V':
			if(name.substring(1, 3).equals("is")) {
				if(name.substring(3).equals("Innocent"))
					if(victim==null)
						return 1;
					else
						return 0;
				else
					throw new RuntimeException("Strawberry: unrecognised variable - \"" + name + "\"");
			} else {
				/*returns 0 if doens't exist*/
				return vt.getIntProperty(killer, name.substring(1));
			}
			
		/*function*/
		case 'f':
			if(name.substring(1,4).equals("min"))
				return min(name.substring(5), killer, victim, vt);
			if(name.substring(1,4).equals("max"))
				return min(name.substring(5), killer, victim, vt);
			/*else*/	
			throw new RuntimeException("Strawberry: unrecognised variable - \"" + name + "\"");

		/*otherwise*/
		default:
			throw new RuntimeException("Strawberry: unrecognised variable - \"" + name + "\"");
		}
	}

	private List varify(String contents) {
		/*We split on "," and then use Scorer at them*/
		List vars = new ArrayList();
		Stack brackets = new Stack();
		int length = contents.length();
		int start = 0;
		
		try {
			for(int i=0; i<length; i++) {
				char c = contents.charAt(i);
				switch(c) {
					case '{':
						brackets.push("{");
						break;
					case '}':
						if(!((String) brackets.pop()).equals("}"))
							throw new RuntimeException("Strawberry: mismatched bracket in funciton variables");
						break;
					case '(':
						brackets.push("(");
						break;
					case ')':
						if(!((String) brackets.pop()).equals(")"))
							throw new RuntimeException("Strawberry: mismatched bracket in funciton variables");
						break;
	
					case ',':
						if(!brackets.empty())
							break; /*could contain another funciton within this one*/
						vars.add(new Scorer(contents.substring(start,i), 0));
						start=i+1;
						break;
				}
			}
			if(!brackets.empty())
				throw new RuntimeException("Strawberry: unmatched bracket in function variables");
			vars.add(new Scorer(contents.substring(start), 0));
		} catch (ConfigFileException foo) {
			throw new RuntimeException("Strawberry: bad equation while processing function variables");
		}

		return vars;
	}

	private double min(String contents, AssassinViewer killer, 
			AssassinViewer victim, VarTracker vt) 
	{
		List vars=varify(contents);
		double min_val=Double.MAX_VALUE;
		double cur_val;

		Iterator i=vars.iterator();
		while(i.hasNext()) {
			Scorer current = (Scorer) i.next();
			cur_val=current.execute(killer, victim, vt);
			if(cur_val<min_val)
				min_val=cur_val;
		}
		
		return min_val;
	}
	
	private double max(String contents, AssassinViewer killer, 
			AssassinViewer victim, VarTracker vt) 
	{
		List vars=varify(contents);
		double max_val=Double.MIN_VALUE;
		double cur_val;

		Iterator i=vars.iterator();
		while(i.hasNext()) {
			Scorer current = (Scorer) i.next();
			cur_val=current.execute(killer, victim, vt);
			if(cur_val>max_val)
				max_val=cur_val;
		}
		
		return max_val;
	}

}
