/*
 * CUAG AutoUmpire System
 *
 * Copyright (C) 2003/2004 Matthew Johnson, Adam Biltcliffe,
 * Michael Cripps, Martin O'Leary, Edward Allcutt and James Osborn
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * To Contact the authors, please email assassins@srcf.ucam.org
 *
 */
package org.ucam.srcf.assassins.tuttifrutti;

import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Random;

import org.ucam.srcf.assassins.AssassinStateViewer;
import org.ucam.srcf.assassins.AssassinViewer;
import org.ucam.srcf.assassins.Debug;
import org.ucam.srcf.assassins.Plugin;
import org.ucam.srcf.assassins.PluginStateChange;
import org.ucam.srcf.assassins.exceptions.AssassinSQLException;

public class TFAlgorithm implements TargettingAlgorithm {
	class SeedComp implements Comparator {
		public int compare(Object as1, Object as2) {
			AssassinStateViewer asv1 = (AssassinStateViewer) (state.get(as1));
			AssassinStateViewer asv2 = (AssassinStateViewer) (state.get(as2));
			return asv2.getIntProperty(name, "seed") - asv1.getIntProperty(name, "seed");
		}
		public boolean equals(Object as1, Object as2) {
			return (compare(as1,as2) == 0);
		}
	}
	
	private Plugin plugin;
	private TargettingStructure tStruct = new TargettingStructure();

	private Map state;
	private static int numTargets = 3;
	private String name = "tuttifrutti";
	private static Random generator = new Random();


	public String getName() {return name;}

	private static TargettingStructure reallocate(TargettingStructure ts) {
		int playernum = ts.getAssassinList().size();
		if (Debug.debug) Debug.print(TFAlgorithm.class, "Getting matrix");
		byte[][] tmatrix = ts.getTargMatrix();
		if (Debug.debug) Debug.print(TFAlgorithm.class, "Got");
		int[] ntargs = new int[playernum];
		int[] nasses = new int[playernum];
		boolean fail = true;
		byte[][] nmatrix = new byte[playernum][playernum];
		
		while (fail) {
			nmatrix = (byte[][]) tmatrix.clone();
			for (int i = 0; i < playernum; i++) {
				ntargs[i] = 0;
				nasses[i] = 0;
				for (int j = 0; j < playernum; j++) {
					if (nmatrix[i][j] == 1) ntargs[i]++;
					if (nmatrix[i][j] == -1) nasses[i]++;
				}
			}
			fail = false;
			for (int i = 0; i < playernum; i++) {
				int count = 0;
				while ((ntargs[i] < numTargets) && (count < playernum * 20)) {
					int j = generator.nextInt(playernum);
					if ((i != j) && (nasses[j] < numTargets) && (nmatrix[i][j] == 0)) {
						ntargs[i]++;
						nasses[j]++;
						nmatrix[i][j] = 1;
						nmatrix[j][i] = -1;
						count = 0;
					}
					count++;
				}
				if (count > playernum * 20) {
					if (Debug.debug) Debug.print(TFAlgorithm.class, "Failed on "+i);
					fail = true;
				}
			}
		}
		if (Debug.debug) Debug.print(TFAlgorithm.class, "Calculated targetting matrix");
		return new TargettingStructure(nmatrix, ts.getAssassinList());
	}

	public TFAlgorithm(Plugin p, Map state) {
		this.plugin = p;
		this.state = state;
	}

	public List getAssassinList() {
		return tStruct.getAssassinList();
	}

	public List getTargets(AssassinViewer a) {
		return tStruct.getTargets(a);
	}

	public List reallocateTargets() throws AssassinSQLException {
		if (Debug.debug) Debug.print(TFAlgorithm.class, "in reallocateTargets");
		TargettingStructure oldts = (TargettingStructure) tStruct.clone();
		if (Debug.debug) Debug.print(TFAlgorithm.class, "Cloned structure");
		boolean isInitial = tStruct.isEmpty();
		int playernum = tStruct.getAssassinList().size();
		if (Debug.debug) Debug.print(TFAlgorithm.class, "About to reallocate");
		tStruct = reallocate(tStruct);
		if (Debug.debug) Debug.print(TFAlgorithm.class, "Done");
		if (isInitial) {
			System.out.println("Is Initial");
			List alist = tStruct.getAssassinList();
			List tlist = tStruct.triList();
			int t = tlist.size();
			while (t > 0) {
				TargettingStructure trial = (TargettingStructure)tStruct.clone();

				for (Iterator i = alist.iterator(); i.hasNext();) {
					AssassinViewer a = (AssassinViewer) i.next();
					for (Iterator j = alist.iterator(); j.hasNext();) {
						AssassinViewer b = (AssassinViewer) j.next();
						if (0 == generator.nextInt(5 * playernum)) {
							trial.takeTarget(a,b);
							trial.takeTarget(b,a);
						}
					}
				}
				for (Iterator i = tlist.iterator(); i.hasNext();) {
					AssassinViewer a = (AssassinViewer) i.next();
					for (Iterator j = tlist.iterator(); j.hasNext();) {
						AssassinViewer b = (AssassinViewer) j.next();
						if (0 == generator.nextInt(2 * t)) {
							trial.takeTarget(a,b);
							trial.takeTarget(b,a);
						}
					}
				}
				trial = reallocate(trial);
				List tlist2 = trial.triList();
				if (tlist2.size() <= t) {
					tStruct = trial;
					tlist = tlist2;
					t = tlist2.size();
				}
			}
			if (Debug.debug) Debug.print(TFAlgorithm.class, "Girth 4 achieved (or died trying)");
			int[][] distance = tStruct.getFloydMatrix();
			List seedlist = new LinkedList();
			seedlist.add(new Integer(0));
			while (seedlist.size() < playernum) {
				int maxshiny = -1000000;
				int testseed = -1;
				for (int i = 0; i < playernum; i++) {
					int shininess = 1;
					for (Iterator x = seedlist.iterator(); x.hasNext();) {
						int j = ((Integer)x.next()).intValue();
						shininess *= distance[i][j] - 1;
					}
					if ((shininess > maxshiny) && !seedlist.contains(new Integer(i))) {
						maxshiny = shininess;
						testseed = i;
					}
				}
				seedlist.add(new Integer(testseed));
			}

			if (Debug.debug) Debug.print(TFAlgorithm.class, "Got " + seedlist.size() + " seeds");
			List alist2 = new LinkedList(alist);
			Collections.sort(alist2, new SeedComp());
			for (Iterator myiter = alist2.iterator(); myiter.hasNext();) {
				AssassinViewer av = (AssassinViewer) myiter.next();
				System.out.println(av.getPlayerViewer().getName() + " " + ((AssassinStateViewer) state.get(av)).getIntProperty(name, "seed"));
			}
			List alist3 = new LinkedList();
			for (int i = 0; i < playernum; i++) {
				if (Debug.debug) Debug.print(TFAlgorithm.class, "Adding player "+i);
				alist3.add(alist2.get(seedlist.indexOf(new Integer(i))));
			}
			tStruct = new TargettingStructure(tStruct.getTargMatrix(), alist3);
		}
//		List alist = tStruct.getAssassinList();
		List sclist = new LinkedList();
		Iterator i = oldts.getAssassinList().iterator();
		while (i.hasNext()) {
			AssassinViewer a = (AssassinViewer) i.next();
			Iterator j = oldts.getAssassinList().iterator();
			while (j.hasNext()) {
				AssassinViewer b = (AssassinViewer) j.next();
				if (oldts.hasTarget(a,b) && !(tStruct.hasTarget(a,b))) {
					sclist.add(new PluginStateChange(a,plugin,"detarget",b));
				}
				if (!(oldts.hasTarget(a,b)) && tStruct.hasTarget(a,b)) {
					sclist.add(new PluginStateChange(a,plugin,"entarget",b));
				}
			}
		}
		return sclist;	
	}


	public void giveTarget(AssassinViewer ass, AssassinViewer tar) {
		tStruct.giveTarget(ass, tar);
	}
	public void takeTarget(AssassinViewer ass, AssassinViewer tar) {
			tStruct.takeTarget(ass, tar);
	}

	public void removeAssassin(AssassinViewer a) {
		tStruct.removeAssassin(a);
	}


	public void addAssassin(AssassinViewer a) {
		tStruct.addAssassin(a);
	}

}
