/* 
 * CUAG AutoUmpire System
 * 
 * Copyright (C) 2003/2004 Matthew Johnson, Adam Biltcliffe,
 * Michael Cripps, Martin O'Leary, Edward Allcutt and James Osborn
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * To Contact the authors, please email assassins@srcf.ucam.org
 *
 */

package org.ucam.srcf.assassins;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Arrays;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.Vector;

import org.ucam.srcf.assassins.exceptions.AssassinSQLException;
import org.ucam.srcf.assassins.exceptions.NotSerialisedEntityException;
import org.ucam.srcf.assassins.exceptions.StateBuilderException;
import org.ucam.srcf.assassins.exceptions.XMLParseException;
import org.ucam.srcf.assassins.xml.Config;
import org.ucam.srcf.assassins.xml.WebPageParser;

import cx.ath.matthew.cgi.CGI;
import cx.ath.matthew.cgi.CGIHeaderSentException;
import cx.ath.matthew.cgi.CheckBox;
import cx.ath.matthew.cgi.DisplayField;
import cx.ath.matthew.cgi.DropDown;
import cx.ath.matthew.cgi.MultipleDropDown;
import cx.ath.matthew.cgi.Field;
import cx.ath.matthew.cgi.HiddenField;
import cx.ath.matthew.cgi.Password;
import cx.ath.matthew.cgi.SubmitButton;
import cx.ath.matthew.cgi.TextArea;
import cx.ath.matthew.cgi.TextField;


/**
 * The main class that manages the CGI interface.
 */
public class umpirecgi  extends CGI
{
   public static final String authentication_file = "/societies/assassins/private/authenticate.txt";
   public static final String[] colleges = new String[] {
                        "Christ's College",
                        "New Hall ",
                        "Churchill College",
                        "Newnham College",
                        "Clare College",
                        "Pembroke College",
                        "Corpus Christi College",
                        "Peterhouse",
                        "Downing College",
                        "Queens' College",
                        "Emmanuel College",
                        "Robinson College",
                        "Fitzwilliam College",
                        "St Catharine's College",
                        "Girton College",
                        "St Edmund's College",
                        "Gonville and Caius College",
                        "St John's College ",
                        "Homerton College",
                        "Selwyn College ",
                        "Hughes Hall",
                        "Sidney Sussex College",
                        "Jesus College",
                        "Trinity College",
                        "King's College",
                        "Trinity Hall",
                        "Lucy Cavendish",
                        "Wolfson College",
                        "Magdalene College",
                        "APU",
                        "Town"};
 
   public static final String[] waterstatuses = new String[] { "No Water", "Water With Care", "Full Water" };
   public static final String[] PLUGINS = new String[] { 
      "org.ucam.srcf.assassins.chocolate.Chocolate",
      "org.ucam.srcf.assassins.peach.Peach",
      "org.ucam.srcf.assassins.pistachio.Pistachio",
      "org.ucam.srcf.assassins.stracciatella.Stracciatella",
      "org.ucam.srcf.assassins.tuttifrutti.Tuttifrutti",
      "org.ucam.srcf.assassins.watermelon.Watermelon"
   };
   private static String rootuser = null;
   public static String getRootUser()
   {
      return rootuser;
   }

   /** These are set after authentication
    */
   public static String authuser = "";
   static String authpw = "";
   static boolean customnowtime = false;


   /**
    * Just call the CGI stuff
    */
   public static void main(String[] args)
   {
      CGI cgi = new umpirecgi();
      cgi.doCGI(args);
   }

   public static void printlink(CGI cgi, String id, String target, String text)
   {
      if (CGI.canGET()) {
         if (customnowtime) {
            String chr;
            if (target.indexOf('?') < 0) chr = "?"; else chr = "&amp;";
            cgi.out("<a href=\""+target+chr+"nowtime="+Tools.formatTime(Tools.getNowTime())+"\">"+text+"</a>");
         } else
            cgi.out("<a href=\""+target+"\">"+text+"</a>");
      } else {
         id = id.replace('-', '_');
         cgi.out("<form name=\""+id+"\" action=\""+target+"\" method=\"post\" style=\"display: inline\">");
         cgi.out("<input type=\"hidden\" name=\"authpw\" value=\""+umpirecgi.authpw+"\"/>");
         cgi.out("<input type=\"hidden\" name=\"authuser\" value=\""+umpirecgi.authuser+"\"/>");
         if (customnowtime)
            cgi.out("<input type=\"hidden\" name=\"nowtime\" value=\""+Tools.formatTime(Tools.getNowTime())+"\"/>");
         cgi.out("<a href=\"javascript:document."+id+".submit()\">"+text+"</a></form>");
      }
   }

   public static void setupCommandMap(Map commands)
   {
      commands.put("show-plugins", new CGICommand() { 
         public void command(CGI cgi, Map plugins, Map params, long nowtime) 
            throws AssassinSQLException, IOException, ParseException { showPlugins(cgi, plugins, params); } 
         public String getName() { return "Show Plugins"; }
         public String getDescription() { return "Lists all the plugins used by this game"; }
         public boolean isPublic() { return false; }
      } );

      commands.put("logout", new CGICommand() { 
         public void command(CGI cgi, Map plugins, Map params, long nowtime) 
            throws AssassinSQLException, IOException, ParseException { logout(cgi, plugins, params); } 
         public String getName() { return "Logout"; }
         public String getDescription() { return "Logout from the umpire and close your session"; }
         public boolean isPublic() { return true; }
      } );

      commands.put("add-player", new CGICommand() { 
         public void command(CGI cgi, Map plugins, Map params, long nowtime) 
            throws AssassinSQLException, IOException, ParseException { addPlayer(cgi, plugins, params, false); } 
         public String getName() { return "Add Player"; }
         public String getDescription() { return "Adds a player"; }
         public boolean isPublic() { return false; }
      } );

      commands.put("change-participant-pseudonym", new CGICommand() { 
         public void command(CGI cgi, Map plugins, Map params, long nowtime) 
            throws AssassinSQLException, IOException, ParseException, StateBuilderException { changeParticipantPseudonym(cgi, plugins, params); } 
         public String getName() { return "Change Participant Pseudonym"; }
         public String getDescription() { return "Changes the pseudonym that an assassin is using in an event"; }
         public boolean isPublic() { return true; }
      } );


      commands.put("add-participant", new CGICommand() { 
         public void command(CGI cgi, Map plugins, Map params, long nowtime) 
            throws AssassinSQLException, IOException, ParseException, StateBuilderException { addParticipant(cgi, plugins, params); } 
         public String getName() { return "Add Participant"; }
         public String getDescription() { return "Adds an assassin as a participant in an existing event"; }
         public boolean isPublic() { return true; }
      } );

      commands.put("add-assassin", new CGICommand() {
         public void command(CGI cgi, Map plugins, Map params, long nowtime) 
            throws AssassinSQLException, IOException, ParseException, StateBuilderException { addAssassin(cgi, null, plugins, params); } 
         public String getName() { return "Add Assassin"; }
         public String getDescription() { return "Give a player another instance in the game"; }
         public boolean isPublic() { return false; }
      } );

      commands.put("add-event", new CGICommand() { 
         public void command(CGI cgi, Map plugins, Map params, long nowtime) 
            throws AssassinSQLException, IOException, ParseException, StateBuilderException {  addEvent(cgi, plugins, params, false); } 
         public String getName() { return "Add Event"; }
         public String getDescription() { return "Adds an event"; }
         public boolean isPublic() { return true; }
      } );      

      commands.put("edit-event", new CGICommand() { 
         public void command(CGI cgi, Map plugins, Map params, long nowtime) 
            throws AssassinSQLException, IOException, ParseException, StateBuilderException {  editEvent(cgi, plugins, params); } 
         public String getName() { return "Edit Event"; }
         public String getDescription() { return "Edits the event details and allows you to add/modify reports and add participants"; }
         public boolean isPublic() { return true; }
      } );      

      commands.put("edit-assassin", new CGICommand() { 
         public void command(CGI cgi, Map plugins, Map params, long nowtime) 
            throws AssassinSQLException, IOException, ParseException, StateBuilderException {  editAssassin(cgi, plugins, params); }
         public String getName() { return "Edit Assassin"; }
         public String getDescription() { return "Modify state values on an assassin, and change a players personal details"; }
         public boolean isPublic() { return false; }
      } );      

      commands.put("edit-report", new CGICommand() { 
         public void command(CGI cgi, Map plugins, Map params, long nowtime) 
         throws AssassinSQLException, IOException, ParseException, StateBuilderException {  editReport(cgi, plugins, params); } 
         public String getName() { return "Edit Report"; }
         public String getDescription() { return "Edit a player's report"; }
         public boolean isPublic() { return true; }
      } );      

      commands.put("add-report", new CGICommand() { 
         public void command(CGI cgi, Map plugins, Map params, long nowtime) 
            throws AssassinSQLException, IOException, ParseException, StateBuilderException {  addReport(cgi, null, plugins, params); } 
         public String getName() { return "Add Report"; }
         public String getDescription() { return "Adds a report to an event"; }
         public boolean isPublic() { return true; }
      } );      

      commands.put("edit-pseudonym", new CGICommand() {
         public void command(CGI cgi, Map plugins, Map params, long nowtime) 
            throws AssassinSQLException, IOException, ParseException, StateBuilderException {  editPseudonym(cgi, plugins, params); } 
         public String getName() { return "Edit Pseudonym"; }
         public String getDescription() { return "Change an existing pseudonym"; }
         public boolean isPublic() { return true; }
      } );      

      commands.put("add-pseudonym", new CGICommand() {
         public void command(CGI cgi, Map plugins, Map params, long nowtime) 
            throws AssassinSQLException, IOException, ParseException, StateBuilderException {  addPseudonym(cgi, plugins, params); } 
         public String getName() { return "Add Pseudonym"; }
         public String getDescription() { return "Give an assassin another pseudonym"; }
         public boolean isPublic() { return true; }
      } );      

      commands.put("add-event-with-report", new CGICommand() { 
         public void command(CGI cgi, Map plugins, Map params, long nowtime) 
            throws AssassinSQLException, IOException, ParseException, StateBuilderException
            {  
               Event e = addEvent(cgi, plugins, params, true);
               if (null != e) addReport(cgi, e, plugins, params);
            } 
         public String getName() { return "Add Event with Report"; }
         public String getDescription() { return "Adds an event and a report on that event together"; }
         public boolean isPublic() { return true; }
      } );

      commands.put("add-player-with-assassin", new CGICommand() { 
         public void command(CGI cgi, Map plugins, Map params, long nowtime) 
            throws AssassinSQLException, IOException, ParseException, StateBuilderException
            {  
               Player p = addPlayer(cgi, plugins, params, true);
               if (null != p) addAssassin(cgi, p, plugins, params);
            } 
         public String getName() { return "Add Player with Assassin"; }
         public String getDescription() { return "Adds a player and associated assassin at the same time"; }
         public boolean isPublic() { return false; }
      } );

      commands.put("dump-state", new CGICommand() { 
         public void command(CGI cgi, Map plugins, Map params, long nowtime) 
            throws AssassinSQLException, IOException, ParseException, StateBuilderException {  dumpState(cgi, plugins, params); }
         public String getName() { return "Dump State"; }
         public String getDescription() { return "Dumps the current state values for all assassins"; }
         public boolean isPublic() { return false; }
      } );      

      commands.put("update-state", new CGICommand() { 
         public void command(CGI cgi, Map plugins, Map params, long nowtime) 
            throws AssassinSQLException, IOException, ParseException, StateBuilderException {  updateState(cgi, plugins, params); }
         public String getName() { return "Update State"; }
         public String getDescription() { return "Allows all the plugins to check their state and ask any questions they need to"; }
         public boolean isPublic() { return false; }
      } );      

      commands.put("generate-mail", new CGICommand() { 
         public void command(CGI cgi, Map plugins, Map params, long nowtime) 
            throws AssassinSQLException, IOException, ParseException, StateBuilderException {  generateMail(cgi, plugins, params); }
         public String getName() { return "Generate Mail"; }
         public String getDescription() { return "Generates emails to be sent to the players if they need it"; }
         public boolean isPublic() { return false; }
      } );      

      commands.put("generate-pages", new CGICommand() { 
         public void command(CGI cgi, Map plugins, Map params, long nowtime) 
            throws AssassinSQLException, IOException, ParseException, StateBuilderException {  generatePages(cgi, plugins, params); } 
         public String getName() { return "Generate Pages"; }
         public String getDescription() { return "Generates the news and player list pages"; }
         public boolean isPublic() { return false; }
      } );
   }

   /**
    * Turn a Question into an HTML Field. 
    * Also includes plugin identification information
    */
   static Field getPluginField(Plugin p, Question q, int index)
   {
      if (q instanceof OptionQuestion) {
         List l = ((OptionQuestion) q).getOptions();
         if ( l.size() > 0 && l.get(0) instanceof CanonicalEntity)
            return new CanonicalEntityField("::"+p.getPluginName()+"/Option/"+index, q.getLabel(), l, (CanonicalEntity) l.get(((OptionQuestion) q).getDefault()), true);
         if ( l.size() > 0 && l.get(0) instanceof CanonicalViewer)
            return new CanonicalEntityField("::"+p.getPluginName()+"/Option/"+index, q.getLabel(), l, (CanonicalViewer) l.get(((OptionQuestion) q).getDefault()), true);
         else
            return new DropDown("::"+p.getPluginName()+"/Option/"+index, q.getLabel(), ((OptionQuestion) q).getOptions(), ((OptionQuestion) q).getDefault(), true);
      }
      else if (q instanceof TextQuestion)
         return new TextField("::"+p.getPluginName()+"/Text/"+index, q.getLabel(), ((TextQuestion) q).getDefault());
      else if (q instanceof YesNoQuestion)
         return new CheckBox("::"+p.getPluginName()+"/Check/"+index, q.getLabel(), ((YesNoQuestion) q).getDefault());
      else return null;
   }

   static final void editPseudonym(CGI cgi, Map plugins, Map params) throws AssassinSQLException, IOException, ParseException, StateBuilderException
   { 
      Debug.printMap(umpirecgi.class, params);
      if ("2".equals(params.get("pass")) || (null != params.get("1") && !"".equals(params.get("1")))) {
         HTMLForm form = new HTMLForm("umpire.cgi?edit-pseudonym", "Edit");
         Pseudonym p = (Pseudonym) params.get("pseudonym");
         if (null == p) p = Pseudonym.getByPsid(Integer.parseInt((String) params.get("1")));
         if (Debug.debug) Debug.print(umpirecgi.class, "pseudonym: "+p);
         cgi.out("<h3>Edit Pseudonym</h3>");
         form.addField(new HiddenField("pass","3"));
         form.addField(new HiddenField("pseudonym",""+p));
         form.addField(new TextField("value","Pseudonym:",Tools.escapeXML(p.getValue())));
         form.display(cgi);
      } else if ("3".equals(params.get("pass"))) {
         try { cgi.header("Refresh", "2; URL=umpire.cgi"); } catch (CGIHeaderSentException CGIHSe) {} 
         cgi.out("<h3>Edit Pseudonym</h3>");
         cgi.out("<p>");
         cgi.out("  Pseudonym Updated");
         cgi.out("</p>");
         Pseudonym pseudonym = (Pseudonym) params.get("pseudonym");
         String value = (String) params.get("value");
         pseudonym.setValue(value);
         Pseudonym.commitAll();
      } else {
         HTMLForm form = new HTMLForm("umpire.cgi?edit-pseudonym", "Edit");
         cgi.out("<h3>Edit a Pseudonym</h3>");
         cgi.out("<p>");
         cgi.out("  Select Pseudonym to edit:");
         cgi.out("</p>");
         Event.getAllEventsWithHeadline();
         form.addField(new CanonicalEntityField("pseudonym","Pseudonym:",Pseudonym.getAllPseudonyms(),(CanonicalEntity) null));
         form.addField(new HiddenField("pass","2"));
         form.display(cgi);
      }
   }

   static final void editReport(CGI cgi, Map plugins, Map params) throws AssassinSQLException, IOException, ParseException, StateBuilderException
   { 
      Debug.printMap(umpirecgi.class, params);
      if ("2".equals(params.get("pass")) || (null != params.get("1") && !"".equals(params.get("1")))) {
         HTMLForm form = new HTMLForm("umpire.cgi?edit-report", "Edit");
         Report r = (Report) params.get("report");
         if (null == r) r = Report.getByRid(Integer.parseInt((String) params.get("1")));
         if (Debug.debug) Debug.print(umpirecgi.class, "report: "+r);
         LightweightStateBuilder lsb = new LightweightStateBuilder(Tools.getNowTime());
         Assassin a = r.getAssassin();
         Map myState = null;
         if (null != a) {
            lsb.registerInterest(a);
            try {
               myState = lsb.generateState();
            } catch (StateBuilderException sbe) {
               throw new RuntimeException();
            }
         }
         else
            myState = new HashMap();
         cgi.out("<h3>Edit Report</h3>");
         cgi.out("<p>");
         if (a != null)
            cgi.out("  Report by "+a.getPlayer().getName()+":");
         else
            cgi.out("  Comment:");
         cgi.out("</p>");
         form.addField(new HiddenField("pass","3"));
         form.addField(new HiddenField("report",""+r));
         if (a != null) {
            Pseudonym p = r.getPseudonym();
            form.addField(new CanonicalEntityField("pseudonym","Choose Pseudonym:",((AssassinState) myState.get(r.getAssassin())).getPseudonyms(), p));
            form.addField(new CheckBox("nopseud", "Use Real Name", null == p));
         }
         form.addField(new TextArea("text","Report:",r.getText(), 80, 25));
         form.display(cgi);
      } else if ("3".equals(params.get("pass"))) {
         // Done asking questions, so update event
         cgi.out("<h3>Edit Report</h3>");
         cgi.out("<p>");
         cgi.out("  Report Updated");
         cgi.out("</p>");
         Pseudonym pseudonym = (Pseudonym) params.get("pseudonym");
         Report r = (Report) params.get("report");
         String text = (String) params.get("text");
          if ("on".equals(params.get("nopseud")))
             r.setPseudonym(null);
          else
             r.setPseudonym(pseudonym);
         r.setText(text);
         try { cgi.header("Refresh", "2; URL=umpire.cgi"); } catch (CGIHeaderSentException CGIHSe) {} 
         Report.commitAll();
      } else {
         HTMLForm form = new HTMLForm("umpire.cgi?edit-report", "Edit");
         cgi.out("<h3>Edit a report</h3>");
         cgi.out("<p>");
         cgi.out("  Select Report to edit:");
         cgi.out("</p>");
         Event.getAllEventsWithHeadline();
         form.addField(new CanonicalEntityField("report","Report:",Report.getAllReports(),(CanonicalEntity) null));
         form.addField(new HiddenField("pass","2"));
         form.display(cgi);
      }
   }

   static final void editEvent(CGI cgi, Map plugins, Map params) throws AssassinSQLException, IOException, ParseException, StateBuilderException
   {
      if (null != params.get("update")) {
         Event e = (Event) params.get("event");
         e.setEventTime(Tools.unFormatTime((String) params.get("time")));
         e.setHeadline((String) params.get("headline"));
         e.setActive("on".equals(params.get("active")) ? true : false);
         e.setVisible("on".equals(params.get("visible")) ? true : false);
         Event.commitAll();
         cgi.out("<p>");
         cgi.out("   Event updated.");
         cgi.out("</p>");
         try { cgi.header("Refresh", "2; URL=umpire.cgi?edit-event,"+e.getEid()); } catch (CGIHeaderSentException CGIHSe) {} 
      } 
      else {
         Event e = (Event) params.get("event");
         if (null == e) {
            String s = (String) params.get("1");
            if (null != s && !"".equals(s)) {
               int eid = Integer.parseInt(s);
               if ("public".equals(params.get("2"))) eid -= Config.getConfigInt("vanilla","obscurity-offset"); 
               e = Event.getByID(eid);
            }
         }
         if (null == e) {
            cgi.out("<p>");
            cgi.out("   Select the Event to edit.");
            cgi.out("</p>");
            HTMLForm form = new HTMLForm("umpire.cgi?edit-event", "Edit");
            form.addField(new CanonicalEntityField("event","Event:",Event.getAllEventsWithHeadline(),e));
            form.display(cgi);
         } 
         else {
            cgi.out("<p>");
            cgi.out("   Edit the event details, and click below to alter participants and reports.");
            cgi.out("</p>");
            HTMLForm form = new HTMLForm("umpire.cgi?edit-event", "Edit");
            form.addField(new TextField("headline", "Headline: ", e.getHeadline(), 60));
            form.addField(new TextField("time", "Event Time: ", Tools.formatTime(e.getEventTime())));
            form.addField(new CheckBox("active", "Active: ", e.isActive()));
            form.addField(new CheckBox("visible", "Visible: ", e.isVisible()));
            form.addField(new HiddenField("event", ""+e));
            form.addField(new HiddenField("update", "true"));
            form.display(cgi);
            cgi.out("<h4>Participants</h4>");
            cgi.out("<ul>");
            Iterator i = e.getParticipants().keySet().iterator();
            while (i.hasNext()) {
               int n = ((Integer) i.next()).intValue();
               Assassin a = e.getParticipant(n);
               Pseudonym p = e.getParticipantPseudonym(n);
               cgi.out(" <li>"+n+". "+a.getPrintable()+" as "+(null==p?a.getPlayer().getName():p.getPrintable())+" [");
               printlink(cgi, "editps"+n, "umpire.cgi?change-participant-pseudonym,"+e.getEid()+","+n,"Change");
               cgi.out("]</li>");
            }
            cgi.out(" <li>");
            printlink(cgi, "addp", "umpire.cgi?add-participant,"+e.getEid(), "Add Participant");    
            cgi.out(" </li>");
            cgi.out("</ul>");
            cgi.out("<h4>Reports</h4>");
            cgi.out("<ul>");
            i = e.getReports().iterator();
            while (i.hasNext()) {
               Report r = (Report) i.next();
               cgi.out(" <li>"+r.getPrintable()+" [");
               printlink(cgi, "rep"+r.getRid(), "umpire.cgi?edit-report,"+r.getRid(), "Edit");    
               cgi.out("]</li>");
            }
            cgi.out(" <li>");
            printlink(cgi, "addr", "umpire.cgi?add-report,"+e.getEid(), "Add New Report");    
            cgi.out(" </li>");
            cgi.out("</ul>");
         }
      }
   }

   static final void changeParticipantPseudonym(CGI cgi, Map plugins, Map params) throws AssassinSQLException, IOException, ParseException, StateBuilderException
   {
      // 0 = select event, 1 = select participant, 2 = select pseudonym, 3 = commit
      int pass = 0;
      Event e = null;
      Assassin a = null;
      Pseudonym p = null;
      if (null == params.get("pass")) {
         if (null == params.get("1"))
            pass = 0;
         else {
            e = Event.getByID(Integer.parseInt((String) params.get("1")));
            if (null == params.get("2"))
               pass = 1;
            else {
               a = e.getParticipant(Integer.parseInt((String) params.get("2")));
               pass = 2;
            }
         }
      } else {
         pass = Integer.parseInt((String) params.get("pass"));
         if (null != params.get("event")) e = (Event) params.get("event");
         if (null != params.get("assassin")) a = (Assassin) params.get("assassin");
         if (null != params.get("pseudonym")) p = (Pseudonym) params.get("pseudonym");
      }

      cgi.out("<h3>Change Pseudonym used in an Event</h3>");
      HTMLForm form = new HTMLForm("umpire.cgi?change-participant-pseudonym", "Change Pseudonym");

      switch (pass) {
         case 0:
            cgi.out("<p>");
            cgi.out("  Select Event:");
            cgi.out("</p>");
            form.addField(new CanonicalEntityField("event","Event:",Event.getAllEventsWithHeadline(),(CanonicalEntity) null));
            form.addField(new HiddenField("pass","1"));
            form.display(cgi);
            break;
         case 1:
            cgi.out("<p>");
            cgi.out("  Select Participant:");
            cgi.out("</p>");
            form.addField(new DisplayField("Event:", e.getPrintable()));
            form.addField(new HiddenField("event",""+e));
            form.addField(new CanonicalEntityField("assassin","Participant:",e.getParticipants().values(),(CanonicalEntity)null));
            form.addField(new HiddenField("pass","2"));
            form.display(cgi);
            break;
         case 2:
            cgi.out("<p>");
            cgi.out("  Select Pseudonym:");
            cgi.out("</p>");
            form.addField(new DisplayField("Event:", e.getPrintable()));
            form.addField(new HiddenField("event",""+e));
            form.addField(new DisplayField("Participant:",a.getPrintable()));

            LightweightStateBuilder lsb = new LightweightStateBuilder(Tools.getNowTime());
            lsb.registerInterest(a);
            Map myState = null;
            try {
               myState = lsb.generateState();
            } catch (StateBuilderException sbe) {
               throw new RuntimeException();
            }
            form.addField(new HiddenField("assassin",""+a));
            int n = 0;
            Iterator i = e.getParticipants().keySet().iterator();
            while (i.hasNext()) {
               n = ((Integer) i.next()).intValue();
               if (e.getParticipant(n).equals(a)) break;
            }
            p = e.getParticipantPseudonym(n);
            form.addField(new CanonicalEntityField("pseudonym","Pseudonym:",((AssassinState) myState.get(a)).getPseudonyms(), p));
            form.addField(new CheckBox("nopseudonym", "Use Real Name:", null==p));
            form.addField(new HiddenField("pass","3"));
            form.display(cgi);
            break;
         case 3:
            try { cgi.header("Refresh", "2; URL=umpire.cgi?edit-event,"+e.getEid()); } catch (CGIHeaderSentException CGIHSe) {} 
            cgi.out("<p>");
            cgi.out("  Pseudonym Successfully Changed.");
            cgi.out("</p>");
            n = 0;
            i = e.getParticipants().keySet().iterator();
            while (i.hasNext()) {
               n = ((Integer) i.next()).intValue();
               if (e.getParticipant(n).equals(a)) break;
            }
            if ("on".equals(params.get("nopseudonym")))
               e.setParticipantPseudonym(n, null);
            else
               e.setParticipantPseudonym(n, p);
            Event.commitAll();
            break;
      }
   }

   static final void addParticipant(CGI cgi, Map plugins, Map params) throws AssassinSQLException, IOException, ParseException, StateBuilderException
   {
       if ("2".equals(params.get("pass"))) {
          HTMLForm form = new HTMLForm("umpire.cgi?add-participant", "Add");
          Assassin a = (Assassin) params.get("assassin");
          Event e = (Event) params.get("event");
          if (Debug.debug) Debug.print(umpirecgi.class, "assassin: "+a+", event: "+e);
          LightweightStateBuilder lsb = new LightweightStateBuilder(Tools.getNowTime());
          lsb.registerInterest(a);
          Map myState = null;
          try {
             myState = lsb.generateState();
          } catch (StateBuilderException sbe) {
             throw new RuntimeException();
          }
          cgi.out("<h3>Add Participant to Event</h3>");
          cgi.out("<p>");
          cgi.out("  Select Pseudonym for Participant:");
          cgi.out("</p>");
          form.addField(new HiddenField("pass","3"));
          form.addField(new HiddenField("assassin",""+a));
          form.addField(new HiddenField("event",""+e));
          if (Debug.debug) Debug.print(umpirecgi.class, "state: "+myState.get(a));
          if (Debug.debug) Debug.print(umpirecgi.class, "pseudonyms: "+((AssassinState) myState.get(a)).getPseudonyms());
          form.addField(new CanonicalEntityField("pseudonym","Pseudonym:",((AssassinState) myState.get(a)).getPseudonyms(),(CanonicalEntity) null));
          form.addField(new CheckBox("nopseud", "Use Real Name", false));
          form.display(cgi);
       } else if ("3".equals(params.get("pass"))) {
          // Done asking questions, so update event
          cgi.out("<h3>Add Participant to Event</h3>");
          cgi.out("<p>");
          cgi.out("  Participant Successfully Added");
          cgi.out("</p>");
          Pseudonym pseudonym = (Pseudonym) params.get("pseudonym");
          Assassin a = (Assassin) params.get("assassin");
          Event e = (Event) params.get("event");
          try { cgi.header("Refresh", "2; URL=umpire.cgi?edit-event,"+e.getEid()); } catch (CGIHeaderSentException CGIHSe) {} 
          int m = e.getMaxParticipantID();
          if ("on".equals(params.get("nopseud")))
             e.setParticipant(++m, a, null);
          else
             e.setParticipant(++m, a, pseudonym);
          Event.commitAll();
       } else {
          HTMLForm form = new HTMLForm("umpire.cgi?add-participant", "Add");
          cgi.out("<h3>Add a participant in an Event</h3>");
          cgi.out("<p>");
          cgi.out("  Select Event to add participant to and assassin to add:");
          cgi.out("</p>");
          if (null != params.get("1") && !"".equals(params.get("1"))) {
             Event e = Event.getByID(Integer.parseInt((String) params.get("1")));
             form.addField(new DisplayField("Event:", e.getPrintable()));
             form.addField(new HiddenField("event",""+e));
          }
          else
             form.addField(new CanonicalEntityField("event","Event:",Event.getAllEventsWithHeadline(),(CanonicalEntity) null));
          form.addField(new CanonicalEntityField("assassin","Participant:",Assassin.getAllAssassinsOrderByRealName(),(CanonicalEntity)null));
          form.addField(new HiddenField("pass","2"));
          form.display(cgi);
       }
   }

   static final void editAssassin(CGI cgi, Map plugins, Map params) throws AssassinSQLException, IOException, ParseException, StateBuilderException
   {
      if (null != params.get("update")) {

         Assassin a = (Assassin) params.get("assassin");
         LightweightStateBuilder lsb = new LightweightStateBuilder(Tools.getNowTime());
         lsb.registerInterest(a);
         Map state=null;
         try {
            state = lsb.generateState();
         } catch (StateBuilderException e1) {
            e1.printStackTrace();
            throw new RuntimeException();
         }
         AssassinState ast = (AssassinState) state.get(a);
         Player pl = a.getPlayer();

         Event e = new Event();
         e.setActive(true);
         e.setEventTime(Tools.getNowTime());

         pl.setName((String) params.get("name"));
         pl.setEmail((String) params.get("email"));
         pl.setCollege((String) params.get("college"));
         pl.setAddress((String) params.get("address"));
         pl.setWaterStatus((String) params.get("water"));
         pl.setNotes((String) params.get("notes"));

         // resurrect them
         if ("on".equals((String) params.get("alive")) && !ast.isAlive())
            e.addStateChange(new Resurrection(a));
         // kill them
         if (!"on".equals((String) params.get("alive")) && ast.isAlive())
            e.addStateChange(new Death(a, a));
         // deactivate them
         if (!"on".equals((String) params.get("active")) && ast.isActive())
            e.addStateChange(new AssassinDeactivation(a));
         // activate them
         if ("on".equals((String) params.get("active")) && !ast.isActive())
            e.addStateChange(new AssassinActivation(a));

         // plugin details
         List plugs = ast.getKnownPlugins();
         Iterator i = plugs.iterator();
         while (i.hasNext()) {
            String p = (String) i.next();
            if (Debug.debug) Debug.print(umpirecgi.class, "got "+p);
            if (!"vanilla".equals(p)) {
               Map m = ast.getState(p);
               Iterator j = m.keySet().iterator();
               while (j.hasNext()) {
                  String key = (String) j.next();
                  if (Debug.debug) Debug.print(umpirecgi.class, "got "+key);
                  Object cval = m.get(key);

                  // string
                  if (cval instanceof String) 
                     if (!cval.equals((String) params.get(p+"_"+key))) {
                        StateChange sc = new PluginStateChange(a,(Plugin) plugins.get(p), 
                              key, (String) params.get(p+"_"+key));
                        e.addStateChange(sc);
                     }

                  // CanonicalEntity
                  else if (cval instanceof CanonicalEntity) {
                        StateChange sc = new PluginStateChange(a,(Plugin) plugins.get(p), 
                              key, (CanonicalEntity) params.get(p+"_"+key));
                        e.addStateChange(sc);
                  }
               }
            }
         }

         cgi.out("<p>");
         cgi.out("   Assassin updated.");
         cgi.out("</p>");
         try { cgi.header("Refresh", "2; URL=umpire.cgi?edit-assassin,"+a.getAid()); } catch (CGIHeaderSentException CGIHSe) {} 
      } 
      else if (null != params.get("setproperty")) {
         Assassin a = (Assassin) params.get("assassin");
         Event e = new Event();
         e.setActive(true);
         e.setEventTime(Tools.getNowTime());
         StateChange sc = new PluginStateChange(a,(Plugin) plugins.get(params.get("plugin")), 
               (String) params.get("variable"), (String) params.get("value"));
         e.addStateChange(sc);
         cgi.out("<p>");
         cgi.out("   Property Added. ");
         cgi.out("</p>");
         try { cgi.header("Refresh", "2; URL=umpire.cgi?edit-assassin%2C"+a.getAid()); } catch (CGIHeaderSentException CGIHSe) {} 
      }
      else {
         Assassin a = (Assassin) params.get("assassin");
         if (null == a) {
            String s = (String) params.get("1");
            if (null != s && !"".equals(s))
               a = Assassin.getByAid(Integer.parseInt(s));
         }
         if (null == a) {
            cgi.out("<p>");
            cgi.out("   Select the Assassin to edit.");
            cgi.out("</p>");
            HTMLForm form = new HTMLForm("umpire.cgi?edit-assassin", "Edit");
            form.addField(new CanonicalEntityField("assassin","Assassin:",Assassin.getAllAssassinsOrderByRealName(),a));
            form.display(cgi);
         } 
         else {
            cgi.out("<p>");
            cgi.out("   Edit the assassin details.");
            cgi.out("</p>");
            LightweightStateBuilder lsb = new LightweightStateBuilder(Tools.getNowTime());
            lsb.registerInterest(a);
            Map state=null;
            try {
               state = lsb.generateState();
            } catch (StateBuilderException e1) {
               e1.printStackTrace();
               throw new RuntimeException();
            }
            AssassinState ast = (AssassinState) state.get(a);
            Player pl = a.getPlayer();
            HTMLForm form = new HTMLForm("umpire.cgi?edit-assassin", "Edit");
            form.addField(new TextField("name", "Real Name:", pl.getName()));
            form.addField(new TextField("email", "Email:", pl.getEmail()));
            java.util.Arrays.sort(colleges);
            form.addField(new DropDown("college", "College:", colleges, pl.getCollege(), false));
            form.addField(new TextField("address", "Address:", pl.getAddress()));
            form.addField(new DropDown("water", "Water Status:", waterstatuses, pl.getWaterStatus(), false));
            form.addField(new TextArea("notes", "Other Notes (will be sent to assassins):", pl.getNotes()));
            form.addField(new CheckBox("active", "Active: ", ast.isActive()));
            form.addField(new CheckBox("alive", "Alive: ", ast.isAlive()));
            form.addField(new HiddenField("assassin", ""+a));
            form.addField(new HiddenField("update", "true"));
            // plugin details
            List plugs = ast.getKnownPlugins();
            Iterator i = plugs.iterator();
            while (i.hasNext()) {
               String p = (String) i.next();
               if (Debug.debug) Debug.print(umpirecgi.class, "got "+p);
               if (!"vanilla".equals(p)) {
                  form.addField(new DisplayField("Plugin", "<b>"+p+"</b>"));
                  Map m = ast.getState(p);
                  Iterator j = m.keySet().iterator();
                  while (j.hasNext()) {
                     String key = (String) j.next();
                     if (Debug.debug) Debug.print(umpirecgi.class, "got "+key);
                     Object cval = m.get(key);

                     // string
                     if (cval instanceof String)
                        form.addField(new TextField(p+"_"+key ,key ,(String) cval));

                     // CanonicalEntity
                     else if (cval instanceof CanonicalEntity) {
                        form.addField(new CanonicalEntityField(p+"_"+key,
                                 key, CanonicalEntity.getAll((CanonicalEntity) cval), (CanonicalEntity) cval));
                     }
                  }
               }
            }
            form.display(cgi);
            
            cgi.out("<p class='center'>");
            cgi.out("   Manually add a plugin value to an assassin:");
            cgi.out("</p>");
            form = new HTMLForm("umpire.cgi?edit-assassin", "Manual Add");
            form.addField(new DropDown("plugin", "Plugin:", new Vector(plugins.keySet()), "", false));
            form.addField(new TextField("variable", "Field:", ""));
            form.addField(new TextField("value", "Value:", ""));
            form.addField(new HiddenField("setproperty", "true"));
            form.addField(new HiddenField("assassin", ""+a));
            form.display(cgi);
            cgi.out("<h4>Pseudonyms</h4>");
            cgi.out("<ul>");
            i = ast.getPseudonyms().iterator();
            while (i.hasNext()) {
               Pseudonym ps = (Pseudonym) i.next();
               cgi.out(" <li>"+Tools.escapeXML(ps.getValue())+" [");
               printlink(cgi, "ps"+ps.getPsid(), "umpire.cgi?edit-pseudonym,"+ps.getPsid(),"Edit");    
               cgi.out("]</li>");
            }
            cgi.out(" <li>");
            printlink(cgi, "addp", "umpire.cgi?add-pseudonym,"+a.getAid(), "Add Pseudonym");    
            cgi.out(" </li>");
            cgi.out("</ul>");
         }
      }
   }

   static final void updateState(CGI cgi, Map plugins, Map params) throws AssassinSQLException, IOException, ParseException, StateBuilderException
   {
      cgi.out("<p>");
      cgi.out("   Calculating state, please wait....");
      cgi.out("</p>");
      cgi.flush();

      if (Debug.debug) Debug.print(umpirecgi.class, "Building State...");
      // build our state
      StateBuilder sb = new StateBuilder(plugins);
      Map state;
      try {
         state = sb.buildStateAtTime(Tools.getNowTime());
      } catch (StateBuilderException sbe) {
         throw new RuntimeException();
      }
      Map statev = viewerifyStateMap(state);

      if (Debug.debug) Debug.print(umpirecgi.class, "Get Event Parameter");
      // get the event we are using
      Event e = (Event) params.get("event");
      if (null == e) {
         e = new Event();
         e.setEventTime(Tools.getNowTime());
         e.setModifiedTime(Tools.getNowTime());
         e.setActive(true);
         e.setVisible(false);
      }

      if (Debug.debug) Debug.print(umpirecgi.class, "Building answers...");
      // build the vector of questions we have asked
      Map oqs = new HashMap();
      boolean haveqs = false;
      Iterator i = plugins.values().iterator();
      while (i.hasNext()) {
         Plugin p = (Plugin) i.next();
         oqs.put(p, new Vector());
         Vector qs = p.updateState(statev, (Vector) oqs.get(p), (EventViewer) e.getViewer()); 
         if (null != qs) {
            oqs.put(p, qs);
            haveqs = true;
         }
      }

      Map nqs;
      if (null != params.get("pass")) {
         // ::chocolate/Check/0
         i = params.keySet().iterator();
         while (i.hasNext()) {
            String param = (String) i.next();
            if (!param.startsWith("::")) continue;
            int a = param.indexOf("/");
            Plugin p = (Plugin) plugins.get(param.substring(2,a));
            int b = param.indexOf("/", a+1);
            String type = param.substring(a+1, b);
            int index = Integer.parseInt(param.substring(b+1));
            if (Debug.debug) Debug.print(umpirecgi.class, "Got a question for plugin "+p+" type "+type+" index "+index+" value "+params.get(param));
            Vector qs = (Vector) oqs.get(p);
            Question question = (Question) qs.get(index);
            if ("Check".equals(type)) {
               if ("on".equals(params.get(param)))
                  ((YesNoQuestion) question).setAnswer(true);
               else
                  ((YesNoQuestion) question).setAnswer(false);
            }
            else if ("Text".equals(type)) {
               ((TextQuestion) question).setAnswer((String) params.get(param));
            }
            else if ("Option".equals(type)) {
               ((OptionQuestion) question).setAnswer(
                           ((OptionQuestion) question).getOptions().indexOf(params.get(param)));
            }
         }

         if (Debug.debug) Debug.print(umpirecgi.class, "Asking plugins for questions");
         // generate a list of questions
         nqs = new HashMap();
         haveqs = false;
         i = plugins.values().iterator();
         while (i.hasNext()) {
            Plugin p = (Plugin) i.next();
            if (Debug.debug) Debug.print(umpirecgi.class, "Updating for plugin: "+p);
            Vector qs = p.updateState(statev, (Vector) oqs.get(p), (EventViewer) e.getViewer()); 
            if (Debug.debug) Debug.print(umpirecgi.class, "Got questions: "+qs);
            if (null != qs) {
               nqs.put(p, qs);
               haveqs = true;
            }
         }
      }
      else {
         nqs = oqs;
      }
      
      // ask all the questions
      if (haveqs) {
         cgi.out("<p>");
         cgi.out("   Plugins have questions to ask:");
         cgi.out("</p>");
         HTMLForm form = new HTMLForm("umpire.cgi?update-state", "Submit");
         form.addField(new HiddenField("event",""+e));
         form.addField(new HiddenField("pass","true"));
         i = plugins.values().iterator();
         while (i.hasNext()) {
            Plugin p = (Plugin) i.next();
            Vector qs = (Vector) nqs.get(p);
            if (null == qs) continue;
            Iterator j = qs.iterator();
            for (int k = 0; j.hasNext(); k++) {
               form.addField(umpirecgi.getPluginField(p, (Question) j.next(), k));
            }
         }
         form.display(cgi);   
      }
      else {
         cgi.out("<p>");
         cgi.out("   All questions asked, state updated.");
         cgi.out("</p>");
         try { cgi.header("Refresh", "2; URL=umpire.cgi"); } catch (CGIHeaderSentException CGIHSe) {} 
      }
   }

   static String sendEmails(Map state, Map plugins, boolean force, String manualText) throws AssassinSQLException, ParseException
   {
      String emails = "";
      String from = Config.getConfig("vanilla", "from-address");
      Iterator i = state.values().iterator();
      Map statev = viewerifyStateMap(state);
      while(i.hasNext())
      {
         AssassinState as = (AssassinState) i.next();
         if (!as.isAlive()) continue;
         if (!as.isActive()) continue;
         boolean needsemail = force;
         Iterator j = plugins.values().iterator();
         AssassinStateViewer asv = new AssassinStateViewer(as);
         while(j.hasNext() && !needsemail) 
            needsemail |= ((Plugin) j.next()).playerNeedsEmail(asv);


         if (needsemail) {
            Player p = as.getAssassin().getPlayer();
            // get text
            String emailText = "Your Details:\n";
            emailText += "           Name: "+p.getName()+"\n";
            emailText += "        Address: "+p.getAddress()+"\n";
            emailText += "        College: "+p.getCollege()+"\n";
            emailText += "   Water Status: "+p.getWaterStatus()+"\n";
            emailText += "          Notes: "+p.getNotes()+"\n";

            j = plugins.keySet().iterator();
            while(j.hasNext()) {
               String pname = (String) j.next();
               Plugin pl = (Plugin) plugins.get(pname);
               String emailadd = pl.generateEmail(asv, statev);
               if (null != emailadd && !"".equals(emailadd)) emailText += "\n----------------\n" + emailadd;
            }

            // print text & headers
            emails += "MAIL FROM:" + from + "\n";
            emails += "RCPT TO:" + p.getEmail() + "\n";
            emails += "DATA\n";
            emails += emailText + "\n";
            emails += manualText + "\n\n.\n";
         }
      }
      return emails;
   }

   static final void dumpState(CGI cgi, Map plugins, Map params) throws AssassinSQLException , StateBuilderException
   {
      StateBuilder sb = new StateBuilder(plugins);
      Map state;
      try {
         state = sb.buildStateAtTime(Tools.getNowTime());
      } catch (StateBuilderException sbe) {
         throw new RuntimeException();
      }
      Iterator i = state.values().iterator();
      while (i.hasNext())
      {
         AssassinState s = (AssassinState) i.next();
         Assassin a = s.getAssassin();
         Player p = a.getPlayer();
         cgi.out("<h3>Assassin played by "+p.getName()+"</h3>");
         cgi.out("<table>");
         cgi.out("<thead>");
         cgi.out("  <tr><td colspan='3'>Properties</td></tr>");
         cgi.out("  <tr>");
         cgi.out("     <td>Plugin</td>");
         cgi.out("     <td>Field</td>");
         cgi.out("     <td>Value</td>");
         cgi.out("  </tr>");
         cgi.out("</thead>");
         cgi.out("<tbody>");
         Iterator pni = s.getKnownPlugins().iterator();
         while (pni.hasNext())
         {
            String pname = (String) pni.next();
            Map pstate = s.getState(pname);
            Iterator ki = pstate.keySet().iterator();
            while (ki.hasNext())
            {
               String field = (String) ki.next();
               Object v = s.getProperty(pname, field);
               cgi.out("  <tr>");
               cgi.out("     <td>"+pname+"</td>");
               cgi.out("     <td>"+field+"</td>");
               cgi.out("     <td>"+(null == v ? "null" : v.toString())+"</td>");
               cgi.out("  </tr>");
            }
         }
         cgi.out("</tbody>");
         cgi.out("</table>");
         cgi.out("<h4>Pseudonyms:</h4>");
         cgi.out("<ul>");
         Iterator psi = s.getPseudonyms().iterator();
         while (psi.hasNext())
         {
            Pseudonym ps = (Pseudonym) psi.next();
            cgi.out("   <li>"+ps.getValue()+"</li>");
         }
         cgi.out("</ul>");
      }
   }

   static final Player addPlayer(CGI cgi, Map plugins, Map params, boolean withassassin) throws AssassinSQLException, IOException
   {
      if (params.get("submit") != null) {
         Player p = new Player();
         p.setName((String) params.get("name"));
         p.setEmail((String) params.get("email"));
         p.setCollege((String) params.get("college"));
         p.setAddress((String) params.get("address"));
         p.setWaterStatus((String) params.get("water"));
         p.setNotes((String) params.get("notes"));
         Player.commitAll();
         cgi.out("<p>");
         cgi.out("  New player added successfully.");
         cgi.out("</p>");
         if (!withassassin) try { cgi.header("Refresh", "2; URL=umpire.cgi"); } catch (CGIHeaderSentException CGIHSe) {} 
         return p;
      }

      else {
         HTMLForm form;

         if (withassassin) form = new HTMLForm("umpire.cgi?add-player-with-assassin", "Add");
         else form = new HTMLForm("umpire.cgi?add-player", "Add");

         form.addField(new TextField("name", "Real Name:", ""));
         form.addField(new TextField("email", "Email:", "@cam.ac.uk"));
         /* Sort the colleges into alphabetical order */
         java.util.Arrays.sort(colleges);
         form.addField(new DropDown("college", "College:", colleges, "", false));
         form.addField(new TextArea("address", "Address:", ""));
         form.addField(new DropDown("water", "Water Status:", waterstatuses, "", false));
         form.addField(new TextArea("notes", "Notes (will be sent to assassins):", ""));
         cgi.out("<h3>Add a new player</h3>");
         cgi.out("<p>");
         cgi.out("  Please enter their details below:");
         cgi.out("</p>");
         form.display(cgi);
         return null;
      }
   }

   static final void addAssassin(CGI cgi, Player p, Map plugins, Map params) throws AssassinSQLException, IOException, StateBuilderException
   {
      // pass two, do plugin questions
      if ("2".equals(params.get("pass"))) {
         p = (Player) params.get("player");
         Event e = new Event();
         e.setActive(true);
         e.setEventTime(Tools.getNowTime());
         Assassin a = new Assassin();
         a.setPlayer(p);
         if ("on".equals(params.get("active")))
            e.addStateChange(new AssassinActivation(a));
         String s = (String) params.get("pseudonym");
         if (null != s && !"".equals(s)) {
            Pseudonym ps = new Pseudonym();
            ps.setValue(s);
            ps.setAssassin(a);
            e.addStateChange(new PseudonymGain(a, ps));
         }
         Event.commitAll();
         Assassin.commitAll();

         cgi.out("<h3>Add a new assassin</h3>");
         cgi.out("<p>");
         cgi.out("   Assassin Created, Additional Questions:");
         cgi.out("</p>");

         HTMLForm form = new HTMLForm("umpire.cgi?add-assassin", "Add");
         form.addField(new HiddenField("pass", "3"));

         Iterator plgs = plugins.values().iterator();
         if (Debug.debug) Debug.print(umpirecgi.class, "assassin: "+a);
         LightweightStateBuilder lsb = new LightweightStateBuilder(Tools.getNowTime());
         lsb.registerInterest(a);
         Map myState=null;
         try {
            myState = lsb.generateState();
         } catch (StateBuilderException e1) {
            e1.printStackTrace();
            throw new RuntimeException();
         }
         Map statev = viewerifyStateMap(myState);

         while (plgs.hasNext()) {
            Plugin pl = (Plugin) plgs.next();
            // EventViewer ev = (EventViewer) e.getViewer();
            AssassinViewer av = (AssassinViewer) a.getViewer();
            Vector qs = pl.getQuestions(av, statev);
            Iterator qi = qs.iterator();
            for (int i=0; qi.hasNext(); i++)
               form.addField(umpirecgi.getPluginField(pl, (Question) qi.next(), i));
         }
         Assassin.commitAll();
         Event.commitAll();

         form.addField(new HiddenField("assassin", ""+a));
         form.addField(new HiddenField("event", ""+e));
         form.display(cgi);
      }

      // pass 3, commit everything
      else if ("3".equals(params.get("pass"))) {
         // parse out plugin fields from POST
         Iterator plgs = plugins.values().iterator();
         Assassin a = (Assassin) params.get("assassin");
         LightweightStateBuilder lsb = new LightweightStateBuilder(Tools.getNowTime());
         lsb.registerInterest(a);
         Map myState=null;
         try {
            myState = lsb.generateState();
         } catch (StateBuilderException e1) {
            e1.printStackTrace();
            throw new RuntimeException();
         }
         Map statev = viewerifyStateMap(myState);
         AssassinViewer av = (AssassinViewer) a.getViewer();
         EventViewer ev = (EventViewer) ((Event) params.get("event")).getViewer();

         while (plgs.hasNext()) {
            Plugin pl = (Plugin) plgs.next();
            Vector qs = pl.getQuestions(av, statev);
            if (null == qs) continue;
            Iterator qi = qs.iterator();
            for (int i=0; qi.hasNext(); i++) {
               Question q = (Question) qi.next();
               String s = "";
               if (q instanceof TextQuestion) s = "/Text/";
               else if (q instanceof YesNoQuestion) s = "/Check/";
               else if (q instanceof OptionQuestion) s = "/Option/";
               s = "::"+pl.getPluginName()+s+i;
               String par = (String)params.get(s);
               if (null == par) continue;
               if (q instanceof TextQuestion) 
                  ((TextQuestion) q).setAnswer(par);
               else if (q instanceof YesNoQuestion) 
                  if ("on".equals(par))
                     ((YesNoQuestion) q).setAnswer(true);
                  else
                     ((YesNoQuestion) q).setAnswer(false);
               else if (q instanceof OptionQuestion) 
                  ((OptionQuestion) q).setAnswer(Integer.parseInt(par));
            }
            pl.returnQuestions(av, ev, statev, qs);
         }

         Event.commitAll();

         cgi.out("<p>");
         cgi.out("  New assassin added successfully, additional questions completed.");
         cgi.out("</p>");
         try { cgi.header("Refresh", "2; URL=umpire.cgi?edit-assassin,"+a.getAid()); } catch (CGIHeaderSentException CGIHSe) {} 
      }

      // pass one, do vanilla questions
      else {
         HTMLForm form = new HTMLForm("umpire.cgi?add-assassin", "Add");
         if (null == p) {
            if (Debug.debug) Debug.printMap(umpirecgi.class, params);
            if (null != params.get("1") && !"".equals(params.get("1"))){
               p = Player.getByPid( Integer.parseInt((String) params.get("1")) );
               form.addField(new DisplayField("Player:", p.getPrintable()));
               form.addField(new HiddenField("player",""+p));
            } else {
               List ps = Player.getAllPlayersOrderByName();
               form.addField(new CanonicalEntityField("player",
                        "Select Player to add assassin for:", ps, p));
            }
         } else  {
            form.addField(new DisplayField("Player:", p.getPrintable()));
            form.addField(new HiddenField("player", p.toString()));
         }
         form.addField(new CheckBox("active", "Make assassin active?", true));
         form.addField(new TextField("pseudonym", "Initial Pseudonym:", ""));
         form.addField(new HiddenField("pass", "2"));
         cgi.out("<h3>Add a new assassin</h3>");
         form.display(cgi);
      }
   }

   static final Event addEvent(CGI cgi, Map plugins, Map params, boolean withreport) throws AssassinSQLException, IOException, ParseException , StateBuilderException
   { 
      // By Mike - fixx0red by matt
      if ("3".equals(params.get("pass"))){ // pass 3
         Event e = new Event();
         Iterator i = params.keySet().iterator();
         while (i.hasNext()) {
            String key = (String) i.next();
            if (key.startsWith("assassin")) {
               String idx = key.substring(8);
               Assassin a = (Assassin) params.get(key);
               Pseudonym p = (Pseudonym) params.get("pseudonym"+idx);
               int id = Integer.parseInt(idx) + 1;
               if ("on".equals(params.get("nopseud"+idx)))
                  e.setParticipant(id, a, null);
               else
                  e.setParticipant(id, a, p);
            }
         }
         e.setEventTime(Tools.unFormatTime((String)params.get("event-time")));
         e.setHeadline((String) params.get("headline"));
         e.setActive("on".equals(params.get("active")) ? true : false);
         e.setVisible("on".equals(params.get("visible")) ? true : false);
         cgi.out("<h3>Add a new event</h3>");
         cgi.out("<p>");
         cgi.out("Add Death");
         cgi.out("</p>");
         HTMLForm form;
         if (withreport) 
            form = new HTMLForm("umpire.cgi?add-event-with-report", "No More Deaths");
         else
            form = new HTMLForm("umpire.cgi?add-event", "No More Deaths");
         form.addField(new HiddenField("pass","4"));
         form.addField(new HiddenField("index","0"));
         form.addField(new CanonicalEntityField("victim","Select victim:",e.getParticipants().values(),(CanonicalEntity) null));
         form.addField(new CanonicalEntityField("killer","Select killer:",e.getParticipants().values(),(CanonicalEntity) null));
         Event.commitAll();
         form.addField(new HiddenField("event",""+e));
         form.addField(new SubmitButton("submit","Add Death"));
         form.display(cgi); 
         return null;
      } else if ("4".equals(params.get("pass")))  { // pass 4
         if ("Add Death".equals(params.get("submit"))) {
            // If we want to add another death
            HTMLForm form;
            if (withreport) 
               form = new HTMLForm("umpire.cgi?add-event-with-report", "No More Deaths");
            else
               form = new HTMLForm("umpire.cgi?add-event", "No More Deaths");
            Event e = (Event) params.get("event");
            form.addField(new HiddenField("event",""+e));
            form.addField(new HiddenField("pass","4"));

            Assassin victim = (Assassin) params.get("victim");
            Assassin killer = (Assassin) params.get("killer");
            Death d = new Death(killer,victim);
            e.addStateChange(d);
            Event.commitAll();
            cgi.out("<h3>Add a new event</h3>");
            cgi.out("<p>");
            cgi.out("Added kill of "+victim.getPrintable()+" by "+killer.getPrintable()+".");
            cgi.out("</p>");

            cgi.out("<p>");
            cgi.out("Add Death, Submit Form to continue without adding another death");
            cgi.out("</p>");
            form.addField(new CanonicalEntityField("victim","Select victim:",e.getParticipants().values(),(CanonicalEntity) null));
            form.addField(new CanonicalEntityField("killer","Select killer:",e.getParticipants().values(),(CanonicalEntity) null));
            form.addField(new SubmitButton("submit","Add Death"));
            form.display(cgi); 
         } else {
            // Done adding deaths, let's ask plugins
            Event e = (Event) params.get("event");

            cgi.out("<h3>Add a new event</h3>");
            cgi.out("<p>");
            cgi.out("   Event added, Additional Questions:");
            cgi.out("</p>");

            HTMLForm form;
            if (withreport) 
               form = new HTMLForm("umpire.cgi?add-event-with-report", "Answer");
            else
               form = new HTMLForm("umpire.cgi?add-event", "Answer");
            form.addField(new HiddenField("pass", "5"));

            Iterator plgs = plugins.values().iterator();
            LightweightStateBuilder lsb = new LightweightStateBuilder(Tools.getNowTime());
            for (Iterator i = e.getParticipants().values().iterator(); i.hasNext();) {
               lsb.registerInterest((Assassin)i.next());
            }
            Map statev = null;
            try {
               statev = viewerifyStateMap(lsb.generateState());
            } catch (StateBuilderException sbe) {
               throw new RuntimeException();
            }
            while (plgs.hasNext()) {
               Plugin pl = (Plugin) plgs.next();
               EventViewer ev = (EventViewer) e.getViewer();
               Vector qs = pl.getQuestions(ev, statev);
               Iterator qi = qs.iterator();
               for (int i=0; qi.hasNext(); i++)
                  form.addField(umpirecgi.getPluginField(pl, (Question) qi.next(), i));
            }
            Event.commitAll();

            form.addField(new HiddenField("event", ""+e));
            form.display(cgi);
         }
         return null;
      } else if ("5".equals(params.get("pass"))) {// pass 5

         Event e = (Event) params.get("event");
         Iterator ps = plugins.values().iterator();
         LightweightStateBuilder lsb = new LightweightStateBuilder(Tools.getNowTime());
         for (Iterator i = e.getParticipants().values().iterator(); i.hasNext();) {
            lsb.registerInterest((Assassin)i.next());
         }
         Map statev = null;
         try {
            statev = viewerifyStateMap(lsb.generateState());
         } catch (StateBuilderException sbe) {
            throw new RuntimeException();
         }
         while (ps.hasNext()) {
            Plugin p = (Plugin) ps.next();
            EventViewer ev = (EventViewer) e.getViewer();
            Vector qs = p.getQuestions(ev, statev);
            if (null == qs) continue;
            Iterator qi = qs.iterator();
            for (int i2=0; qi.hasNext(); i2++) {
               Question q = (Question) qi.next();
               String s = "";
               if (q instanceof TextQuestion) s = "/Text/";
               else if (q instanceof YesNoQuestion) s = "/Check/";
               else if (q instanceof OptionQuestion) s = "/Option/";
               s = "::"+p.getPluginName()+s+i2;
               String par = (String)params.get(s);
               if (null == par) continue;
               if (q instanceof TextQuestion) 
                  ((TextQuestion) q).setAnswer(par);
               else if (q instanceof YesNoQuestion) 
                  if ("on".equals(par))
                     ((YesNoQuestion) q).setAnswer(true);
                  else
                     ((YesNoQuestion) q).setAnswer(false);
               else if (q instanceof OptionQuestion) 
                  ((OptionQuestion) q).setAnswer(Integer.parseInt(par));
            }
            p.returnQuestions(ev, statev, qs);
         }
         Event.commitAll();

         cgi.out("<h3>Add a new event</h3>");
         cgi.out("<p>");
         cgi.out("  New event added successfully. All statechange events added. All questions asked.");
         cgi.out("</p>");
         if (!withreport) try { cgi.header("Refresh", "2; URL=umpire.cgi?edit-event,"+e.getEid()); } catch (CGIHeaderSentException CGIHSe) {} 
         return e;
      } else if ("2".equals(params.get("pass"))) {// pass 2
         HTMLForm form;
         if (withreport) 
            form = new HTMLForm("umpire.cgi?add-event-with-report", "Add");
         else
            form = new HTMLForm("umpire.cgi?add-event", "Add");
         form.addField(new HiddenField("pass","3"));
         Object o = params.get("assassins");
         List as = new LinkedList();
         if (o instanceof List) 
            as = (List) o;
         else if (o instanceof Assassin)
            as.add(o);
         LightweightStateBuilder lsb = new LightweightStateBuilder(Tools.getNowTime());
         for (Iterator i=as.iterator(); i.hasNext();) {
            lsb.registerInterest((Assassin)i.next());
         }
         Map myState = null;
         try {
            myState=lsb.generateState();
         } catch (StateBuilderException sbe) {
            throw new RuntimeException();
         }
         Iterator i = as.iterator();
         cgi.out("<h3>Add a new event</h3>");
         cgi.out("<p>");
         cgi.out("Select Pseudonyms");
         cgi.out("</p>");
         int idx=0;
         while (i.hasNext()) {
            Assassin a;
            a = (Assassin) i.next();
            form.addField(new HiddenField("assassin"+idx,""+a));
            form.addField(new CanonicalEntityField("pseudonym"+idx,(idx+1)+"."+a.getPlayer().getName(),((AssassinState) myState.get(a)).getPseudonyms(),(CanonicalEntity) null));
            form.addField(new CheckBox("nopseud"+idx, "Use Real Name", false));
            ++idx;
         }
         form.addField(new TextField("event-time","Event Time:", Tools.formatTime(Tools.getNowTime())));
         String defheadline = Config.getConfig("vanilla", "default-headline");
         form.addField(new TextField("headline","Headline:", null==defheadline?"":defheadline));
         form.addField(new CheckBox("active","Make Active?", true));
         form.addField(new CheckBox("visible","Make Visible?", true));
         form.display(cgi);
         return null;
      } else {
         // pass 1
         List as = Assassin.getAllAssassinsOrderByRealName();
         if (Debug.debug) Debug.print(umpirecgi.class,  "assassins got: "+as+" ("+as.size()+")");
         cgi.out("<h3>Add a new event</h3>");
         cgi.out("<p>");
         cgi.out("Select the assassins involved in the event: ");
         cgi.out("</p>");
         HTMLForm form;
         if (withreport) 
            form = new HTMLForm("umpire.cgi?add-event-with-report", "Add");
         else
            form = new HTMLForm("umpire.cgi?add-event", "Add");
         form.addField(new MultipleCanonicalEntityField("assassins","",as,(CanonicalEntity) null));
         form.addField(new HiddenField("pass","2"));
         form.display(cgi);
         return null;
      }
   }

   static final void addPseudonym(CGI cgi, Map plugins, Map params) throws AssassinSQLException, IOException , StateBuilderException
   {
      if (null == params.get("submit")) {
         cgi.out("<h3>Add Pseudonym</h3>");
         cgi.out("<p>");
         cgi.out("  Select Assassin and new pseudonym:");
         cgi.out("</p>");
         HTMLForm form = new HTMLForm("umpire.cgi?add-pseudonym", "Add");
         Assassin a = null;
         if (null != params.get("1") && !"".equals(params.get("1"))) {
            a = Assassin.getByAid(Integer.parseInt((String) params.get("1")));
            form.addField(new DisplayField("Assassin:", a.getPrintable()));
            form.addField(new HiddenField("assassin", ""+a));
         } else 
            form.addField(new CanonicalEntityField("assassin","Assassin:",Assassin.getAllAssassinsOrderByRealName(),a));
         form.addField(new TextField("pseudonym", "New Pseudonym:", ""));
         form.display(cgi);
      }
      else {
         cgi.out("<h3>Add Pseudonym</h3>");

         String pseud = (String) params.get("pseudonym");
         if (pseud == null || "".equals(pseud)) {
            cgi.out("<p>");
            cgi.out("Error! Blank Pseudonym. Please go back and try again.");
            cgi.out("</p>");
         } else {
            cgi.out("<p>");
            cgi.out("Pseudonym Added.");
            cgi.out("</p>");
            Assassin a = (Assassin) params.get("assassin");
            try { cgi.header("Refresh", "2; URL=umpire.cgi?edit-assassin,"+a.getAid()); } catch (CGIHeaderSentException CGIHSe) {} 
            Pseudonym p = new Pseudonym();
            p.setValue(pseud);
            Event e = new Event();
            e.setActive(true);
            e.setEventTime(Tools.getNowTime());
            e.addStateChange(new PseudonymGain(a, p));
         }
      }
   }

   static final void addReport(CGI cgi, Event e, Map plugins, Map params) throws AssassinSQLException, IOException , StateBuilderException
   {
      if ("2".equals(params.get("pass"))) {
         HTMLForm form = new HTMLForm("umpire.cgi?add-report", "Add");
         Assassin a = (Assassin) params.get("assassin");
         e = (Event) params.get("event");
         String comment = (String) params.get("comment");
         if (Debug.debug) Debug.print(umpirecgi.class, "assassin: "+a+", event: "+e);
         if (!"on".equals(comment)) {
            LightweightStateBuilder lsb = new LightweightStateBuilder(Tools.getNowTime());
            lsb.registerInterest(a);
            Map myState = null;
            try {
               myState = lsb.generateState();
            } catch (StateBuilderException sbe) {
               throw new RuntimeException();
            }
            cgi.out("<h3>Add Report</h3>");
            cgi.out("<p>");
            cgi.out("  Select Pseudonym for Assassin:");
            cgi.out("</p>");
            form.addField(new HiddenField("assassin",""+a));
            if (Debug.debug) Debug.print(umpirecgi.class, "state: "+myState.get(a));
            if (Debug.debug) Debug.print(umpirecgi.class, "pseudonyms: "+((AssassinState) myState.get(a)).getPseudonyms());
            form.addField(new CanonicalEntityField("pseudonym","",((AssassinState) myState.get(a)).getPseudonyms(),(CanonicalEntity) null));
            form.addField(new CheckBox("nopseud", "Use Real Name", false));
         }
         else {
            cgi.out("<h3>Add Report</h3>");
            cgi.out("<p>");
            cgi.out("  Adding Comment:");
            cgi.out("</p>");
         }
         form.addField(new HiddenField("pass","3"));
         form.addField(new HiddenField("event",""+e));
         form.addField(new TextArea("report","Report:","Report Text goes here", 80, 25));
         form.display(cgi);
      } else if ("3".equals(params.get("pass"))) {
         // Done asking questions, so update event
         cgi.out("<h3>Add Report</h3>");
         cgi.out("<p>");
         cgi.out("  Report/Comment Successfully Added");
         cgi.out("</p>");
         Pseudonym pseudonym = (Pseudonym) params.get("pseudonym");
         Assassin a = (Assassin) params.get("assassin");
         String report = (String) params.get("report");
         e = (Event) params.get("event");
         Report r = new Report();
         r.setAssassin(a);
         if ("on".equals(params.get("nopseud")))
            r.setPseudonym(null);
         else
            r.setPseudonym(pseudonym);
         r.setText(report);
         e.addReport(r);
         Event.commitAll();
         try { cgi.header("Refresh", "2; URL=umpire.cgi?edit-event,"+e.getEid()); } catch (CGIHeaderSentException CGIHSe) {} 
      } else {
         HTMLForm form = new HTMLForm("umpire.cgi?add-report", "Add");
         cgi.out("<h3>Add a report</h3>");
         cgi.out("<p>");
         cgi.out("  Select Event to add report to and reporting assassin:");
         cgi.out("</p>");
         if (null == e && null != params.get("1") && !"".equals(params.get("1"))) {
            e = Event.getByID(Integer.parseInt((String) params.get("1")));
            form.addField(new DisplayField("Event:", e.getPrintable()));
            form.addField(new HiddenField("event",""+e));
            form.addField(new CanonicalEntityField("assassin","Assassin:",new Vector(e.getParticipants().values()),(CanonicalEntity)null));
         }
         else {
            form.addField(new CanonicalEntityField("event","Event:",Event.getAllEventsWithHeadline(),e));
            form.addField(new CanonicalEntityField("assassin","Assassin:",Assassin.getAllAssassinsOrderByRealName(),(CanonicalEntity)null));
         }
         form.addField(new CheckBox("comment","Add Comment:",false));
         form.addField(new HiddenField("pass","2"));
         form.display(cgi);
      }
   }

    static final void generateMail(CGI cgi, Map plugins, Map params)  throws AssassinSQLException, IOException, ParseException , StateBuilderException
    {
       // build the email file
       if ("2".equals(params.get("pass"))) {
          boolean force = false;
          String manualText = "";
          String outFile = null;
          force = params.containsKey("force");
          manualText = (String) params.get("text");
          outFile = (String) params.get("output");

          if (null == outFile) outFile = "email";
          
          cgi.out("<h3>Send Emails to Players</h3>");
          cgi.out("<p>");
          cgi.out("   The following emails will be sent to players, and has been stored in the file '"+outFile+"'.");
          cgi.out("</p>");
          HTMLForm form = new HTMLForm("umpire.cgi?generate-mail", "Send Emails");
          form.addField(new HiddenField("pass", "3"));
          form.addField(new HiddenField("output", outFile));
          form.display(cgi);

          StateBuilder sb = new StateBuilder(plugins);
          Map state;
          try {
             state = sb.buildStateAtTime(Tools.getNowTime());
          } catch (StateBuilderException sbe) {
             throw new RuntimeException();
          }
          String emails = sendEmails(state, plugins, force, manualText);
          File o = new File(outFile);
          PrintWriter out = new PrintWriter(new FileOutputStream(o));
          out.println(emails);
          out.flush();
          out.close();
          cgi.out("<hr/>");
          cgi.out("<pre>");
          cgi.out(emails);
          cgi.out("</pre>");
          cgi.out("<hr/>");
       }
       // send emails
       else if ("3".equals(params.get("pass"))) {
          String outFile = null;
          outFile = (String) params.get("output");
          if (null == outFile) outFile = "email";
          
          Tools.exec(new String[] {
             "/bin/sh",
             "-c",
             "/usr/sbin/sendmail -bS < "+outFile,
             });

          cgi.out("<h3>Send Emails to Players</h3>");
          cgi.out("<p>");
          cgi.out("   Email sent.");
          cgi.out("</p>");          
          try { cgi.header("Refresh", "2; URL=umpire.cgi"); } catch (CGIHeaderSentException CGIHSe) {} 
       }
       // ask about it
       else {
          cgi.out("<h3>Send Emails to Players</h3>");
          cgi.out("<p>");
          cgi.out("   By default this will email all players with outstanding information they need to be sent, such as new targets, "+
                  "password changes or competency warnings.");
          cgi.out("   You can also force an email to go to all players. All emails will contain the player details as a check.");
          cgi.out("</p>");
          
          HTMLForm form = new HTMLForm("umpire.cgi?generate-mail", "Generate Emails");
          form.addField(new TextField("output", "File to save emails to:", "email."+new SimpleDateFormat("yyyyMMdd").format(new Date())));
          form.addField(new CheckBox("force", "Send email to all players:", false));
          form.addField(new TextArea("text", "Extra text to send to players:", ""));
          form.addField(new HiddenField("pass", "2"));
          form.display(cgi);
       }
    }
    
    static final void generatePages(CGI cgi, Map plugins, Map params)  throws AssassinSQLException, IOException, ParseException, StateBuilderException 
    { 
       cgi.out("<h3>Generating News and List Pages</h3>");
       cgi.out("<p>");
       cgi.out("   The news pages and player listings are being generated now...");
       cgi.out("</p>");
       cgi.flush();
       StateBuilder sb = new StateBuilder(plugins);
       Map state = sb.doNews(Tools.getNowTime());
       String path = Config.getConfig("vanilla","path-to-pages");
       cgi.out("<p>");
       cgi.out("   Done. Please remember to check the pages for any mistakes or information leaks.");
       cgi.out("   The pages are available <a href='"+path+"'>here</a>.");
       cgi.out("</p>");
       // need to run xsltproc on all the news....
       String xslt_template = Config.getConfig("vanilla","xsl-file");

       /* Iterating over all the desired templates */
       int i = 1;
       String myTemplate = Config.getConfig("vanilla","template-" + i);
       while (null != myTemplate) {
          String out = Config.getConfig("vanilla","template-page-name-" + i);
          if (Debug.debug) Debug.print(umpirecgi.class, "Path "+path+", Template "+myTemplate+", out "+out);

          // generate the XML
          try {
             WebPageParser wl = new WebPageParser(myTemplate,state,plugins,path+out+".xml",sb.getPages());
             wl.generate();
          } catch (XMLParseException XMLPe) {
             throw new IOException(XMLPe.getMessage());
          }

          // convert to HTML
          if (Debug.debug) Debug.print(umpirecgi.class,"exec(/usr/bin/xsltproc -o "+path+out+".html "+xslt_template+" "+path+out+".xml)");
          Tools.exec(new String[] {
             "/usr/bin/xsltproc",
             "-o",
             path+out+".html",
             xslt_template,
             path+out+".xml"});
          myTemplate = Config.getConfig("vanilla","template-" + ++i);
       }

       return; 
    }

    static final void logout(CGI cgi, Map plugins, Map params)
    {
       try {
          cgi.setcookie("authuser", "");
          cgi.setcookie("authpw", "");
       } catch (CGIHeaderSentException CGIHSe) {}
       cgi.out("<p>");
       cgi.out("  Thank you, "+umpirecgi.authuser+", you have been logged out.");
       cgi.out("</p>");
       return; 
    }

    static final void showPlugins(CGI cgi, Map plugins, Map params)
    {
       cgi.out("<p>");
       cgi.out("  I have instantiated the following plugins:");
       cgi.out("</p>");
       cgi.out("<ul>");
       Iterator i = plugins.keySet().iterator();
       while (i.hasNext()) {
          String pname = (String) i.next();
          cgi.out("   <li>"+pname+"</li>");
       }
       cgi.out("</ul>");
       return; 
    }

    /**
     * Converts a byte array into a hex string
     */
    String toHexStr(byte[] bs)
    {
       String h = "";
       for (int k=0; k<20; k++) {
          int[] a = new int[2];
          a[0] = bs[k] & 0x0F;
          a[1] = (bs[k] >> 4) & 0x0F;
          for (int j=0; j<2; j++) switch (a[j]) {
            case 0x0: h+="0"; break;
            case 0x1: h+="1"; break;
            case 0x2: h+="2"; break;
            case 0x3: h+="3"; break;
            case 0x4: h+="4"; break;
            case 0x5: h+="5"; break;
            case 0x6: h+="6"; break;
            case 0x7: h+="7"; break;
            case 0x8: h+="8"; break;
            case 0x9: h+="9"; break;
            case 0xa: h+="a"; break;
            case 0xb: h+="b"; break;
            case 0xc: h+="c"; break;
            case 0xd: h+="d"; break;
            case 0xe: h+="e"; break;
            case 0xf: h+="f"; break;
          }
       }
       return h;
    }
     
   static Map viewerifyParamMap(Map in)
   {
      Map out = new HashMap();
      Iterator i = in.keySet().iterator();
      while (i.hasNext())
      {
         String key = (String) i.next();
         Object value = in.get(key);
         if (value instanceof String)
            out.put(key, value);
         else if (value instanceof Entity)
            out.put(key, ((Entity) value).getViewer());
         else if (value instanceof List) {
            List l = new LinkedList();
            Iterator j = ((List) value).iterator();
            while (j.hasNext()) {
               Object o = j.next();
               if (o instanceof String)
                  l.add(o);
               else if (o instanceof Entity)
                  l.add(((Entity) o).getViewer());
            }
            out.put(key, l);
         }
      }
      return out;
   }

   static Map viewerifyStateMap(Map in)
   {
      Map out = new HashMap();
      Iterator i = in.keySet().iterator();
      while (i.hasNext())
      {
         Assassin key = (Assassin) i.next();
         AssassinState value = (AssassinState) in.get(key);
         if (Debug.debug)
         {
            if (null == key || null == value)
               Debug.print(umpirecgi.class, "null key or value in viewerifyStateMap");
         }
         AssassinViewer av = new AssassinViewer(key);
         AssassinStateViewer asv = new AssassinStateViewer(value);
      //         if (Debug.debug) Debug.print(this, "inserting into new map: " + av + "," + asv);
         out.put(av, asv);
      }
      return out;
   }

   void printmap(Map a)
   {
      Iterator i = a.keySet().iterator();
      while (i.hasNext()) {
         Object k = i.next();
         Object v = a.get(k);
         out (k + " => " + v + " <br/>");
      }
   }

   /**
    * Checks for a valid authentication, or prompts for it if not.
    */
   boolean authenticate(Map POST, Map GET, Map ENV, Map COOKIE, String[] params) throws CGIHeaderSentException
   {
      if (Debug.debug) {
         Debug.print(this, "Authenticating with:");
         Debug.print(this, "existing: "+umpirecgi.authuser+"/"+umpirecgi.authpw);
         Debug.print(this, "POST: "+POST.get("authuser")+"/"+POST.get("authpw"));
         Debug.print(this, "COOKIE: "+COOKIE.get("authuser")+"/"+COOKIE.get("authpw"));
      }
      if (umpirecgi.authuser != "" && umpirecgi.authpw != "") return true;
    
      // check details in cookies
      if (DatabaseBackend.validateUser((String) COOKIE.get("authuser"), (String) COOKIE.get("authpw"))) {
         umpirecgi.authpw = (String) COOKIE.get("authpw");
         umpirecgi.authuser = (String) COOKIE.get("authuser");
         return true;
      }

      // check details in POST
      if (POST.get("authuser") != null && POST.get("authpw") != null) {
         // TODO: check in here that they are valid. Individal functions should check if username is correct.
         if (DatabaseBackend.validateUser((String) POST.get("authuser"), (String) POST.get("authpw"))) {
            if (POST.get("authlogin") != null) {
               POST.remove("submit");
               POST.remove("authlogin");
            }
            umpirecgi.authpw = (String) POST.get("authpw");
            umpirecgi.authuser = (String) POST.get("authuser");
            String path = (String) ENV.get("REQUEST_URI");
            if (null != path) path = path.substring(0, path.lastIndexOf('/'));
            // set cookie with 8hr timeout
            setcookie("authpw", umpirecgi.authpw, path, (String) ENV.get("SERVER_NAME"), new Date(System.currentTimeMillis() + 28800000), true);
            setcookie("authuser", umpirecgi.authuser, path, (String) ENV.get("SERVER_NAME"), new Date(System.currentTimeMillis() + 28800000), true);
            POST.remove("authuser");
            POST.remove("authpw");
            return true;
         }
         POST.remove("authuser");
         POST.remove("authpw");
         if (POST.get("authlogin") != null) {
            POST.remove("submit");
            POST.remove("authlogin");
         }
      }
      
      // show form
      {
         sendHTMLHeader("Log In");
         out("<p>");
         out("   Please log in.");
         out("</p>");
         String p = "";
         for (int i=0; i<params.length; i++)
            p += params[i]+",";
         HTMLForm form = new HTMLForm("umpire.cgi?"+p, "Log In");
         Iterator i = POST.keySet().iterator();
         while (i.hasNext()) {
            String name = (String) i.next();
            String value = (String) POST.get(name);
            form.addField(new HiddenField(name, value));
         }
         form.addField(new HiddenField("authlogin", "true"));
         form.addField(new TextField("authuser", "Email Address:", authuser));
         form.addField(new Password("authpw", "Password:", authpw));
         form.display(this);
         sendHTMLFooter();
         return false;
      }      
   }

   Map translateEntities(Map input)
   {
      Map output = new HashMap();
      Iterator i = input.keySet().iterator();
      while (i.hasNext()) {
         String k = (String) i.next();
         Object o = input.get(k);
         if (o instanceof List) {
            List l = new LinkedList();
            Iterator j = ((List) o).iterator();
            while (j.hasNext()) {
               String s = (String) j.next();
               Object p = null;
               try {
                  p = CanonicalEntity.deSerialise(s);
               } catch (NotSerialisedEntityException NSEe) {
                  p = s;
               }
               l.add(p);
            }
            output.put(k, l);
         } 
         else if (o instanceof String) {
            String s = (String) o;
            Object p = null;
            try {
               p = CanonicalEntity.deSerialise(s);
            } catch (NotSerialisedEntityException NSEe) {
               p = s;
            }
            output.put(k, p);
         }
      }
      return output;
   }

   public void posttoform(Map POST, HTMLForm form, String[] omit)
   {
      Arrays.sort(omit);
      Iterator i = POST.keySet().iterator();
      while (i.hasNext()) {
         String key = (String) i.next();
         if ("authuser".equals(key)) continue;
         if ("authpw".equals(key)) continue;
         if ("submit".equals(key)) continue;
         if (Arrays.binarySearch(omit, key) < 0) {
            Object o =  POST.get(key);
            if (o instanceof List) {
               Iterator j = ((List) o).iterator();
               while (j.hasNext()) 
                  form.addField(new HiddenField(key, j.next().toString()));
            } else
               form.addField(new HiddenField(key, POST.get(key).toString()));
         }
      }
   }

   /**
    * All the top level stuff in here.
    *
    * Note, atm, ENV and GET are empty.
    */
   protected void cgi(Map POST, Map GET, Map ENV, Map COOKIE, String[] params) throws Exception
   {
      if (Debug.debug) Debug.setOutput("debug.log");
      header("Content-Type", "text/html; charset=UTF-8");
      // try and find a config.xml
      // if it isn't there, run the game setup
      File f = new File("config.xml");
      if (!f.exists()) {
         if (params.length > 0 && params[0].equals("setupgame")) {

            // check auth code
            FileInputStream in = new FileInputStream(authentication_file);
            BufferedReader inp = new BufferedReader(new InputStreamReader(in));
            String authcode = inp.readLine();
            inp.close();

            if (authcode.equals(POST.get("authcode"))) {
               sendHTMLHeader("New Game Setup");
               if ("Done".equals(POST.get("submit"))) {
                  // create game cgi directory and write out config

                  // write config file
                  PrintWriter configfile = new PrintWriter(new FileOutputStream(f));
                  configfile.println("<?xml version=\"1.0\" standalone=\"no\" ?>");
                  configfile.println("<!DOCTYPE game SYSTEM \""+POST.get("xml-files")+"/config.dtd\">");
                  configfile.println("<!-- You Just Lost The ... -->");
                  configfile.println("<game>");
                  configfile.println("   <plugin name=\"vanilla\" class=\"org.ucam.srcf.assassins.Vanilla\">");
                  configfile.println("      <property name=\"game_prefix\">"+POST.get("game_prefix")+"</property>");
                  configfile.println("      <property name=\"dbtype\">"+POST.get("dbtype")+"</property>");
                  configfile.println("      <property name=\"dbhost\">"+POST.get("dbhost")+"</property>");
                  configfile.println("      <property name=\"dbport\">"+POST.get("dbport")+"</property>");
                  configfile.println("      <property name=\"dbname\">"+POST.get("dbname")+"</property>");
                  configfile.println("      <property name=\"dbuser\">"+POST.get("dbuser")+"</property>");
                  configfile.println("      <property name=\"dbpass\">"+POST.get("dbpass")+"</property>");
                  configfile.println("      <property name=\"from-address\">"+POST.get("from-address")+"</property>");
                  configfile.println("      <property name=\"game-start-date\">"+POST.get("game-start-date")+"</property>");
                  configfile.println("      <property name=\"default-headline\">"+POST.get("default-headline")+"</property>");
                  configfile.println("      <property name=\"path-to-pages\">"+POST.get("path-to-pages")+"</property>");
                  configfile.println("      <property name=\"news-page-name\">"+POST.get("news-page-name")+"</property>");
                  configfile.println("      <property name=\"updates-page-name\">"+POST.get("updates-page-name")+"</property>");
                  configfile.println("      <property name=\"headlines-page-name\">"+POST.get("headlines-page-name")+"</property>");
                  configfile.println("      <property name=\"xsl-file\">"+POST.get("xml-files")+"/webpage.xsl</property>");
                  configfile.println("      <property name=\"admin-xsl-file\">"+POST.get("xml-files")+"/adminpage.xsl</property>");
                  configfile.println("      <property name=\"root-address\">"+POST.get("root-address")+"</property>");
                  configfile.println("      <property name=\"email-subject\">"+POST.get("email-subject")+"</property>");
                  configfile.println("      <property name=\"cgiurl\">"+POST.get("cgiurl")+"</property>");
                  Object o = POST.get("cols");
                  if (o != null) {
                     if (o instanceof String) {
                        Vector v = new Vector();
                        v.add(o);
                        o = v;
                     }
                     Iterator i = ((List) o).iterator();
                     while (i.hasNext()) {
                        String colname = (String) i.next();
                        configfile.println("      <property name='colourcount-"+colname+"'>"+
                              POST.get("col."+colname+".count")+"</property>");
                        configfile.println("      <colourfilter name='"+colname+"'>");
                        configfile.println(POST.get("col."+colname+".filter"));
                        configfile.println("      </colourfilter>");
                     }
                  }
                  o = POST.get("templates");
                  if (o != null) {
                     if (o instanceof String) {
                        Vector v = new Vector();
                        v.add(o);
                        o = v;
                     }
                     Iterator i = ((List) o).iterator();
                     int j = 1;
                     while (i.hasNext()) {
                        String tname = (String) i.next();
                        configfile.println("      <property name='template-"+j+"'>"+
                              tname+"</property>");
                        configfile.println("      <property name='template-page-name-"+j+"'>"+
                              POST.get("template."+tname+".output")+"</property>");
                        j++;
                     }
                  }
                  o = POST.get("newsdates");
                  if (o != null) {
                     if (o instanceof String) {
                        Vector v = new Vector();
                        v.add(o);
                        o = v;
                     }
                     Iterator i = ((List) o).iterator();
                     long[] la = new long[((List) o).size()];
                     int j = 0;
                     while (i.hasNext())  {
                        String tstr = (String) i.next();
                        la[j++] = Tools.unFormatTime(tstr);
                     }
                     Arrays.sort(la);
                     for (j--; j >= 0; j--) {
                        configfile.println("      <property name='news-date-"+(j+1)+"'>"+
                              Tools.formatTime(la[j])+"</property>");
                     }
                  }
                  configfile.println("   </plugin>");
                  o = POST.get("plugins");
                  if (o != null) {
                     if (o instanceof String) {
                        Vector v = new Vector();
                        v.add(o);
                        o = v;
                     }
                     Iterator i = ((List) o).iterator();
                     while (i.hasNext()) {
                        String plc = (String) i.next();
                        configfile.println("   <plugin name='"+
                              POST.get(plc+".name")+"' class='"+plc+"'>");
                        Class c = Class.forName(plc);
                        try {
                           List l = (List) 
                              c.getDeclaredMethod("getConfigProperties",
                                    new Class[] {}).invoke(null,new Object[] {});
                           Iterator j = l.iterator();
                           while (j.hasNext()) {
                              ConfigProperty cp = (ConfigProperty) j.next();
                              configfile.println("      <property name='"+
                                    cp.getName()+"'>"+
                                    Tools.escapeXML((String) POST.get(
                                          c.getName()+"."+cp.getName()))+
                                    "</property>");
                           }

                        } catch (Exception e) {}
                        configfile.println("   </plugin>");
                     }
                  }
                  configfile.println("</game>");
                  configfile.close();

                  Tools.exec( new String[] { "chmod", "660", "config.xml" } );

                  // initialise database
                  Config.init("config.xml");
                  DatabaseBackend.init();
                  DatabaseBackend.initTables();

                  // add umpire user
                  Player umpire = new Player();
                  umpire.setPassword((String) POST.get("root-pw"));
                  umpire.setName((String) POST.get("umpire-name"));
                  umpire.setCollege((String) POST.get("umpire-college"));
                  umpire.setEmail((String) POST.get("root-address"));
                  umpire.setAddress((String) POST.get("umpire-address"));
                  /* NOT CREATING ASSASSIN FOR THE UMPIRE
                     Assassin umpass = new Assassin();
                     umpire.addAssassin(umpass);
                     Pseudonym umpps = new Pseudonym();
                     umpps.setValue("The Umpire");
                     Event e = new Event();
                     e.setActive(true);
                     e.setEventTime(Tools.getNowTime());
                     e.addStateChange(new PseudonymGain(umpass, umpps));
                     e.addStateChange(new AssassinActivation(umpass)); */

                  DatabaseBackend.commitAll();

                  out("<p>");
                  out("   The game has been created. Please log in ");
                  printlink(this, "login", "umpire.cgi", "here");
                  out("   .");
                  out("</p>");
               }
               else {
                  HTMLForm form = new HTMLForm("umpire.cgi?setupgame", "Done");
                  posttoform(POST, form, new String[] { "newplugins", "newcol", "newcolcount",
                     "newcolfilter","newtemplatefile", "newtemplatename", "newnewsdate" });
                  if ("Add Plugin".equals(POST.get("submit"))) {
                     Object o = POST.get("newplugins");
                     if (o != null) {
                        if (o instanceof String) { // one new plugin
                           Vector v = new Vector();
                           v.add(o);
                           o = v;
                        }
                        if (o instanceof List) { // several new plugins
                           Iterator i = ((List) o).iterator();
                           while (i.hasNext()) {
                              String pl = (String) i.next();
                              if ("".equals(pl)) continue;

                              Class c = Class.forName(pl);
                              form.addField(new HiddenField("plugins", pl));

                              try {
                                 form.addField(new DisplayField("<b>"+c.getName()+"</b>", 
                                          "<b>"+(String)
                                          c.getDeclaredMethod("getDescription", 
                                             new Class[] {}).invoke(null, 
                                                new Object[] {})+"</b>"));
                              } catch (Exception e) {}

                              String pln = pl.substring(pl.lastIndexOf('.')+1);
                              form.addField(new TextField(pl+".name", 
                                       "Plugin Name:", pln));

                              try {
                                 List l = (List) 
                                    c.getDeclaredMethod("getConfigProperties",
                                          new Class[] {}).invoke(null,new Object[] {});
                                          Iterator j = l.iterator();
                                          while (j.hasNext()) {
                                             ConfigProperty cp = (ConfigProperty) j.next();
                                             form.addField(new TextField(
                                                      c.getName()+"."+cp.getName(), 
                                                      cp.getDescription(), 
                                                      cp.getDefault()));
                                          }
                              } catch (Exception e) {}

                           }
                        }
                     }
                  }
                  else if ("Add Colour Type".equals(POST.get("submit"))) {
                     String colname = (String) POST.get("newcol");
                     form.addField(new DisplayField("<b>Colour Type Added</b>", 
                              colname));
                     form.addField(new HiddenField("cols", colname));
                     form.addField(new HiddenField("col."+colname+".count", 
                              (String) POST.get("newcolcount")));
                     form.addField(new HiddenField("col."+colname+".filter", 
                              (String) POST.get("newcolfilter")));
                  }
                  else if ("Add Template".equals(POST.get("submit"))) {
                     String newtemplate = (String) POST.get("newtemplatefile");
                     form.addField(new DisplayField("<b>Template Added</b>", 
                              newtemplate));
                     form.addField(new HiddenField("templates", newtemplate));
                     form.addField(new HiddenField("template."+newtemplate+".output",
                              (String) POST.get("newtemplatename")));
                  }
                  else if ("Add News Page".equals(POST.get("submit"))) {
                     form.addField(new DisplayField("<b>News Date Added</b>", 
                              (String) POST.get("newnewsdate")));
                     form.addField(new HiddenField("newsdates", 
                              (String) POST.get("newnewsdate")));
                  }
                  form.addField(new MultipleDropDown("newplugins", 
                           "Select Plugins to Add", PLUGINS, null, false));
                  form.addField(new TextField("newplugins", "Add Other Plugin", ""));
                  form.addField(new SubmitButton("submit", "Add Plugin"));
                  form.addField(new TextField("newcol", "Add Colour Type", ""));
                  form.addField(new TextField("newcolcount", "Number of colours", ""));
                  form.addField(new TextArea("newcolfilter", "Colour Filter", "<and>\n	<filter plugin='' property='' condition='' />\n</and>"));
                  form.addField(new SubmitButton("submit", "Add Colour Type"));
                  form.addField(new TextField("newtemplatefile", "Template input file", ""));
                  form.addField(new TextField("newtemplatename", "Template output page name", ""));
                  form.addField(new SubmitButton("submit", "Add Template"));
                  form.addField(new TextField("newnewsdate", "Start date of news page", ""));
                  form.addField(new SubmitButton("submit", "Add News Page"));
                  form.display(this);
               }
            }
            else {
               sendHTMLHeader("Invalid Auth Code");
               out("<p>");
               out("   You have not entered the correct authentication code from "+authentication_file+". You must copy this in exactly to authenticate as the Umpire.");
               out("</p>");
            }
            sendHTMLFooter();
         } else {
            // ask for details of game
            sendHTMLHeader("New Game Setup");
            out("<p>");
            out("   This is the setup system for the Assassins' Guild AutoUmpire System.");
            out("   You will have to enter the authentication code from "+authentication_file+" to create the game.");
            out("</p>");
            PrintWriter authfile = new PrintWriter(new FileOutputStream(authentication_file));
            Random r = new Random(System.currentTimeMillis());
            byte[] authbytes = new byte[20];
            r.nextBytes(authbytes);
            String authstr = toHexStr(authbytes);
            authfile.println(authstr);
            authfile.close();
            HTMLForm form = new HTMLForm("umpire.cgi?setupgame", "Setup");
            form.addField(new TextField("authcode","Authentication Code",""));
            form.addField(new TextField("game_prefix","Game Prefix (e.g. 2004lent)",""));
            form.addField(new TextField("dbtype","Database Type","mysql"));
            form.addField(new TextField("dbhost","Database Host","localhost"));
            form.addField(new TextField("dbport","Database Port","3306"));
            form.addField(new TextField("dbname","Database Name","assassins"));
            form.addField(new TextField("dbuser","Database User","assassins"));
            form.addField(new Password("dbpass","Database Password",""));
            form.addField(new TextField("from-address","Send Emails From","assassins@srcf.ucam.org"));
            form.addField(new TextField("email-subject","Default Email Subject","The Assassins Guild"));
            form.addField(new TextField("game-start-date","Game Start (25-01-2004 05:00:00)",""));
            form.addField(new TextField("default-headline","Default Headline","[P1] killed [P2] ([L2])"));
            form.addField(new TextField("path-to-pages","Location of News Pages",""));
            form.addField(new TextField("cgiurl","URL of CGI Script","/assassins/umpire.cgi"));
            form.addField(new TextField("news-page-name","News Page Stem","news"));
            form.addField(new TextField("updates-page-name","Updates Page Stem","hups"));
            form.addField(new TextField("headlines-page-name","Headlines Page Stem","head"));
            form.addField(new TextField("xml-files","Location of XML files","xml"));
            form.addField(new TextField("umpire-name","Umpire's Name",""));
            form.addField(new TextField("umpire-college","Umpire's College",""));
            form.addField(new TextField("umpire-address","Umpire's Address",""));
            form.addField(new TextField("root-address","Umpire's Email Address",""));
            form.addField(new Password("root-pw","Umpire's Password",""));
            form.display(this);
            sendHTMLFooter();
         }
      }
      // otherwise check the auth tokens
         // check for a parameter
         // no parameter => show command list
         // parameter => do command
      else {
         Map commands = new HashMap();
               
         /*
          * Init config file
          */
         Config.init("config.xml");
         umpirecgi.rootuser = Config.getConfig("vanilla", "root-address");

         /*
          * Init database connection
          */
         DatabaseBackend.init();

         /*
          * Instantiate and init plugins
          */
         Map pluginnames = Config.getPluginList();
         if (Debug.debug) Debug.print(this, "Got "+pluginnames.size()+" plugins");
         Iterator i = pluginnames.keySet().iterator();
         Map plugins = new HashMap();

            while (i.hasNext())
            {
               String pname = (String) i.next();
               String classname = (String) pluginnames.get(pname);
               if (null != classname && !("".equals(classname)))
               {
                  if (Debug.debug) Debug.print(this, "instantiating " + pname);
                  Class c = Class.forName(classname);
                  Plugin p = (Plugin) c.newInstance();
                  if (Debug.debug) Debug.print(this, "setting name");
                  p.setPluginName(pname);
                  if (Debug.debug) Debug.print(this, "setting connector");
                  p.setPluginConnector(new PluginConnector());
                  if (Debug.debug) Debug.print(this, "getting new commands");
                  Map m = p.getCGICommands();                     
                  if (Debug.debug) Debug.printMap(umpirecgi.class, m);
                  if (Debug.debug) Debug.print(this, "Got "+m);
                  commands.putAll(m);
                  plugins.put(pname, p);
               }
            }

         setupCommandMap(commands);

         // TODO: get the time as a parameter?
         {
            if (POST.get("nowtime") != null) {
               customnowtime = true;
               Tools.setNowTime(Tools.unFormatTime((String) POST.get("nowtime")));
            } else if (GET.get("nowtime") != null) {
               customnowtime = true;
               Tools.setNowTime(Tools.unFormatTime((String) GET.get("nowtime")));
            } else
               Tools.setNowTime(new Date().getTime());
         }
         
         if (!authenticate(POST, GET, ENV, COOKIE, params)) return;

         if (Debug.debug) Debug.print(this, "this: '"+authuser+"' root: '"+rootuser+"'");

         if (params.length == 0 && GET.size() != 0) {
            Iterator j = GET.keySet().iterator();
            while (j.hasNext()) {
               String k = (String) j.next();
               if (!"".equals(k) && "".equals(GET.get(k)))
                  params = new String[] { k };
            }
         }
         
         if (params.length == 0) {
            sendHTMLHeader("Select Command");
            out("<p>");
            out("   Select a command from the list below:");
            out("</p>");
            out("<ul class='commandlist'>");
            Object[] a = commands.keySet().toArray();
            Arrays.sort(a);
            for (int j = 0; j < a.length; j++) {
               String command = (String) a[j];
               CGICommand com = (CGICommand) commands.get(command);
               if (!(com.isPublic() || rootuser.equals(authuser))) continue;
               out("   <li>");
               printlink(this, command, "umpire.cgi?"+command, com.getName());
               out("      - "+com.getDescription());
               out("   </li>");
            }
            out("</ul>");
            HTMLForm form = new HTMLForm("umpire.cgi","Run At Time");
            form.addField(new TextField("nowtime","Run at time",
                     Tools.formatTime(Tools.getNowTime())));
            form.display(this);
            sendHTMLFooter();
         }
         else {
            CGICommand com = (CGICommand) commands.get(params[0]);
            if (null == com || !(com.isPublic() || rootuser.equals(authuser))) {
               sendHTMLHeader("Command Not Found");
               out("<p>");
               out("   You have specified an invalid command ("+params[0]+")");
               out("</p>");
               sendHTMLFooter();
            } else {
               sendHTMLHeader(params[0]);
               if (Debug.debug) Debug.print(this, "Running Command "+com);
               Map commandParams = new HashMap();
               if (null != GET) commandParams.putAll(GET);
               if (null != POST) commandParams.putAll(POST);
               if (null != ENV) commandParams.putAll(ENV);
               commandParams = translateEntities(commandParams);
               for (int j=1; j<params.length; j++)
                  commandParams.put(""+j, params[j]);
               com.command(this, plugins, commandParams, Tools.getNowTime());
               DatabaseBackend.commitAll();
               sendHTMLFooter();
            }
         }
      }
   }

   void sendHTMLHeader(String title)
   {
      out("<?xml version=\"1.0\" encoding=\"utf-8\"?>");
      out("<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.1//EN\"");
      out("\"http://www.w3.org/TR/xhtml1/DTD/xhtml11\">");
      out("<html lang=\"en\" xml:lang=\"en\" xmlns=\"http://www.w3.org/1999/xhtml\">");
      out("   <head>");
      out("      <title>Cambridge University Assassins' Guild"+(title==""?"":": "+title)+"</title>");
      out("      <meta http-equiv=\"Content-Type\" content=\"text/html; charset=utf-8\" />");
      out("      <link rel=\"stylesheet\" type=\"text/css\" href=\"assassins.css\" />");
      out("   </head>");
      out("   <body>");
      out("   <hr />");
      out("   <h1>");
      out("      <img class=\"dagger\" src=\"img/cloak2.gif\" alt=\"Cloak &amp; Dagger\" /> ");
      out("      The Assassins' Guild"+(title==""?"":": "+title));
      out("   </h1>");
      out("   <hr />");
      if (customnowtime) {
         out("   <h2>Acting at time: "+Tools.formatTime(Tools.getNowTime())+"</h2>");
         out("   <hr />");
      }
   }

   void sendHTMLFooter()
   {
      String email;
      try {
         email = Config.getConfig("vanilla", "from-address");
      } catch (NullPointerException NPex) { email = ""; }
      out("   <p class=\"center\">");
      printlink(this, "functionlist", "umpire.cgi","Go Back to Function List");
      out("   </p>");
      out("   <hr />   ");
      out("   <p class=\"center\">");
      out("      <a href=\"/assassins/\">Home</a> / <a href=\"mailto:"+email+"\">Email</a>");
      out("      <br />");
      out("      <a href=\"http://validator.w3.org/check/referer\"><img ");
      out("      src=\"img/valid-xhtml11.png\" alt=\"Valid XHTML 1.1\" /></a>");
      out("   </p>");
      out("   <hr />");
      out("   </body>");
      out("</html>");
   }
}

